/*=============================================================================#
 # Copyright (c) 2012, 2021 Original NatTable authors and others.
 # 
 # This program and the accompanying materials are made available under the
 # terms of the Eclipse Public License 2.0 which is available at
 # https://www.eclipse.org/legal/epl-2.0.
 # 
 # SPDX-License-Identifier: EPL-2.0
 # 
 # Contributors:
 #     Original NatTable authors and others - initial API and implementation
 #=============================================================================*/

package org.eclipse.statet.ecommons.waltable.data;

import java.lang.reflect.Method;
import java.util.Locale;


public class ExtendedReflectiveColumnPropertyAccessor<R> extends ReflectiveColumnPropertyAccessor<R> {

	/**
	 * @param propertyNames of the members of the row bean
	 */
	public ExtendedReflectiveColumnPropertyAccessor(final String[] propertyNames) {
		super(propertyNames);
	}

	@Override
	public Object getDataValue(final R rowObj, final long columnIndex) {
		final String propertyName= getColumnProperty(columnIndex);
		if (propertyName.contains(".")) { //$NON-NLS-1$
			return getPropertyValue(rowObj, propertyName);
		}
		else {
			return super.getDataValue(rowObj, columnIndex);
		}
	}

	@Override
	public void setDataValue(final R rowObj, final long columnIndex, final Object newValue) {
		final String propertyName= getColumnProperty(columnIndex);
		if (propertyName.contains(".")) { //$NON-NLS-1$
			setPropertyValue(rowObj, propertyName, newValue);
		}
		else {
			super.setDataValue(rowObj, columnIndex, newValue);
		}
	}

	
	/**
	 * Reads the value of a property out of a given bean via reflection.
	 * @param object the bean out of which the property value should be read
	 * @param propertyName the name of the property which value should be read
	 * @return the property value of the bean
	 */
	private Object getPropertyValue(final Object object, final String propertyName) {
		assert object != null : "object can not be null!"; //$NON-NLS-1$

		String[] propertyChain= null;
		if (propertyName.contains(".")) { //$NON-NLS-1$
			propertyChain= propertyName.split("\\."); //$NON-NLS-1$
		}
		else {
			propertyChain= new String[] {propertyName};
		}

		Object child= object;
		Class<?> objectClass= object.getClass();
		String getterName= null;
		Method getterMethod= null;
		for (final String pc : propertyChain) {
			getterName= "get" + pc.substring(0, 1).toUpperCase(Locale.ROOT) + pc.substring(1); //$NON-NLS-1$
			try {
				getterMethod= objectClass.getMethod(getterName);
				child= getterMethod.invoke(child);
			}
			catch (final NoSuchMethodException e) {
				try {
					getterName= "is" + pc.substring(0, 1).toUpperCase(Locale.ROOT) + pc.substring(1); //$NON-NLS-1$
					getterMethod= objectClass.getMethod(getterName);
					child= getterMethod.invoke(child);
				}
				catch (final Exception e1) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
			catch (final Exception e) {
				e.printStackTrace();
				throw new RuntimeException(e);
			}
			
			if (child != null) {
				objectClass= child.getClass();
			}
			else {
				//null is returned by reflection, therefore we can not go further
				//and null is the correct return value
				break;
			}
		}
		
		return child;
	}

	/**
	 * Sets a value to the property of a bean via reflection.
	 * Also supports dot separated property names to access properties anywhere within the
	 * object graph.
	 * @param object the bean where the property value should be set
	 * @param propertyName the name of the property which should be set
	 * @param value the value to set
	 */
	private void setPropertyValue(final Object object, final String propertyName, final Object value) {
		assert object != null : "object can not be null!"; //$NON-NLS-1$
		
		try {
			Object singlePropertyObject= null;
			String singlePropertyName= null;
			if (propertyName.contains(".")) { //$NON-NLS-1$
				singlePropertyObject= 
					getPropertyValue(object, propertyName.substring(0, propertyName.lastIndexOf("."))); //$NON-NLS-1$
				singlePropertyName= propertyName.substring(propertyName.lastIndexOf(".") + 1); //$NON-NLS-1$
			}
			else {
				singlePropertyObject= object;
				singlePropertyName= propertyName;
			}
			
			
			final String setterName= "set"  //$NON-NLS-1$
				+ singlePropertyName.substring(0, 1).toUpperCase() 
				+ singlePropertyName.substring(1);
			Method setterMethod= null;
			if (value != null) {
				setterMethod= singlePropertyObject.getClass().getMethod(setterName, new Class<?>[] {value.getClass()});
			}
			else {
				//as the value is null we can not access the setter method directly
				//and have to search for the method
				final Method[] methods= singlePropertyObject.getClass().getMethods();
				for (final Method m : methods) {
					if (m.getName().equals(setterName)) {
						setterMethod= m;
					}
				}
			}
			setterMethod.invoke(singlePropertyObject, value);
		} catch (final Exception e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

}
