/*=============================================================================#
 # Copyright (c) 2012, 2021 Original NatTable authors and others.
 # 
 # This program and the accompanying materials are made available under the
 # terms of the Eclipse Public License 2.0 which is available at
 # https://www.eclipse.org/legal/epl-2.0.
 # 
 # SPDX-License-Identifier: EPL-2.0
 # 
 # Contributors:
 #     Original NatTable authors and others - initial API and implementation
 #=============================================================================*/

package org.eclipse.statet.ecommons.waltable.painter.cell.decorator;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;

import org.eclipse.statet.ecommons.waltable.config.IConfigRegistry;
import org.eclipse.statet.ecommons.waltable.coordinate.LRectangle;
import org.eclipse.statet.ecommons.waltable.layer.cell.ILayerCell;
import org.eclipse.statet.ecommons.waltable.painter.cell.CellPainterWrapper;
import org.eclipse.statet.ecommons.waltable.painter.cell.GraphicsUtils;
import org.eclipse.statet.ecommons.waltable.painter.cell.ICellPainter;
import org.eclipse.statet.ecommons.waltable.style.BorderStyle;
import org.eclipse.statet.ecommons.waltable.style.CellStyleAttributes;
import org.eclipse.statet.ecommons.waltable.style.CellStyleUtil;
import org.eclipse.statet.ecommons.waltable.style.IStyle;
import org.eclipse.statet.ecommons.waltable.swt.SWTUtil;


public class LineBorderDecorator extends CellPainterWrapper {

	private final BorderStyle defaultBorderStyle;

	public LineBorderDecorator(final ICellPainter interiorPainter) {
		this(interiorPainter, null);
	}
	
	public LineBorderDecorator(final ICellPainter interiorPainter, final BorderStyle defaultBorderStyle) {
		super(interiorPainter);
		this.defaultBorderStyle= defaultBorderStyle;
	}


	@Override
	public long getPreferredWidth(final ILayerCell cell, final GC gc, final IConfigRegistry configRegistry) {
		final BorderStyle borderStyle= getBorderStyle(cell, configRegistry);
		final long padding= borderStyle != null ? Math.max(borderStyle.getOffset() + borderStyle.getThickness(), 0) : 0;
		
		return super.getPreferredWidth(cell, gc, configRegistry) + (padding * 2);
	}
	
	@Override
	public long getPreferredHeight(final ILayerCell cell, final GC gc, final IConfigRegistry configRegistry) {
		final BorderStyle borderStyle= getBorderStyle(cell, configRegistry);
		final long padding= borderStyle != null ? Math.max(borderStyle.getOffset() + borderStyle.getThickness(), 0) : 0;
		
		return super.getPreferredHeight(cell, gc, configRegistry) + (padding * 2);
	}

	private BorderStyle getBorderStyle(final ILayerCell cell, final IConfigRegistry configRegistry) {
		final IStyle cellStyle= CellStyleUtil.getCellStyle(cell, configRegistry);
		BorderStyle borderStyle= cellStyle.getAttributeValue(CellStyleAttributes.BORDER_STYLE);
		if (borderStyle == null) {
			borderStyle= this.defaultBorderStyle;
		}
		return borderStyle;
	}

	@Override
	public void paintCell(final ILayerCell cell, final GC gc, final LRectangle lRectangle, final IConfigRegistry configRegistry) {
		final BorderStyle borderStyle= getBorderStyle(cell, configRegistry);
		
		final long padding= borderStyle != null ? Math.max(borderStyle.getOffset() + borderStyle.getThickness(), 0) : 0;
		final LRectangle interiorBounds =
			new LRectangle(
					lRectangle.x + padding,
					lRectangle.y + padding,
					lRectangle.width - (padding * 2),
					lRectangle.height - (padding * 2)
			);
		super.paintCell(cell, gc, interiorBounds, configRegistry);
		
		if (borderStyle == null || borderStyle.getThickness() <= 0) {
			return;
		}
		
		// Save GC settings
		final Color originalForeground= gc.getForeground();
		final int originalLineWidth= gc.getLineWidth();
		final int originalLineStyle= gc.getLineStyle();
		
		final long borderOffset= borderStyle.getOffset();
		final int borderThickness= borderStyle.getThickness();
		final LRectangle borderArea= new LRectangle(
						lRectangle.x + borderOffset,
						lRectangle.y + borderOffset,
						lRectangle.width - (borderOffset * 2),
						lRectangle.height - (borderOffset * 2)
				);
		{
			long shift= 0;
			long areaShift= 0;
			if ((borderThickness % 2) == 0) {
				shift= borderThickness / 2;
				areaShift= (shift * 2);
			} else {
				shift= borderThickness / 2;
				areaShift= (shift * 2) + 1;
			}
			borderArea.x+= shift;
			borderArea.y+= shift;
			borderArea.width-= areaShift;
			borderArea.height-= areaShift;
		}

		gc.setLineWidth(borderThickness);
		gc.setLineStyle(SWTUtil.toSWT(borderStyle.getLineStyle()));
		gc.setForeground(borderStyle.getColor());
		gc.drawRectangle(GraphicsUtils.safe(borderArea));

		// Restore GC settings
		gc.setForeground(originalForeground);
		gc.setLineWidth(originalLineWidth);
		gc.setLineStyle(originalLineStyle);
	}
	
}
