/*=============================================================================#
 # Copyright (c) 2007, 2021 Stephan Wahlbrink and others.
 # 
 # This program and the accompanying materials are made available under the
 # terms of the Eclipse Public License 2.0 which is available at
 # https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
 # which is available at https://www.apache.org/licenses/LICENSE-2.0.
 # 
 # SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
 # 
 # Contributors:
 #     Stephan Wahlbrink <sw@wahlbrink.eu> - initial API and implementation
 #=============================================================================*/

package org.eclipse.statet.ecommons.preferences.ui;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.HashMap;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;

import org.eclipse.statet.ecommons.preferences.core.Preference;


/**
 * Provides PropertyChangeSupport (Java Bean) for bean/preferences in
 * a IPreferenceStore (JFace), e.g. OverlayPreferenceStore.
 * 
 * All property-preference-pairs must be registered using {@link #addPreference(String, Preference)}.
 */
public class PreferenceStoreBeanWrapper {
	
	private static class PrefData {
		
		final String beanProperty;
		final Preference converter;
		
		public PrefData(final String beanProperty, final Preference converter) {
			this.beanProperty = beanProperty;
			this.converter = converter;
		}
	}
	
	
	private final IPreferenceStore fStore;
	private final PropertyChangeSupport fBeanSupport;
	
	private final HashMap<String, PrefData> fPreferenceMap= new HashMap<>();
	
	/**
	 * 
	 */
	public PreferenceStoreBeanWrapper(final IPreferenceStore store, final Object bean) {
		fStore = store;
		fBeanSupport = new PropertyChangeSupport(bean);
		
		fStore.addPropertyChangeListener(new IPropertyChangeListener() {
			@Override
			public void propertyChange(final PropertyChangeEvent event) {
				final String storeKey = event.getProperty();
				final PrefData pref = fPreferenceMap.get(storeKey);
				if (pref != null) {
					Object oldValue = event.getOldValue();
					Object newValue = event.getNewValue();
					if (pref.converter != null) {
						if (!pref.converter.getUsageType().isInstance(oldValue)) {
							oldValue = pref.converter.store2Usage((String) oldValue);
						}
						if (!pref.converter.getUsageType().isInstance(newValue)) {
							newValue = pref.converter.store2Usage((String) newValue);
						}
					}
					fBeanSupport.firePropertyChange(pref.beanProperty, 
							oldValue, newValue);
				}
			}
		});
	}
	
	/**
	 * Registers a property/preference, which should be managed.
	 * 
	 * @param beanProperty name of bean property
	 * @param preference identifier for preference store (qualifier not necessary, because not supported by PreferenceStore)
	 */
	public void addPreference(final String beanProperty, final Preference preference) {
		fPreferenceMap.put(preference.getKey(), new PrefData(beanProperty, preference));
	}
	
	public void addPropertyChangeListener(final PropertyChangeListener listener) {
		fBeanSupport.addPropertyChangeListener(listener);
	}
	
	public void addPropertyChangeListener(final String propertyName,
			final PropertyChangeListener listener) {
		fBeanSupport.addPropertyChangeListener(propertyName, listener);
	}
	
	public void removePropertyChangeListener(final PropertyChangeListener listener) {
		fBeanSupport.removePropertyChangeListener(listener);
	}
	
	public void removePropertyChangeListener(final String propertyName,
			final PropertyChangeListener listener) {
		fBeanSupport.removePropertyChangeListener(propertyName, listener);
	}
	
}
