/*=============================================================================#
 # Copyright (c) 2015, 2021 Stephan Wahlbrink and others.
 # 
 # This program and the accompanying materials are made available under the
 # terms of the Eclipse Public License 2.0 which is available at
 # https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
 # which is available at https://www.apache.org/licenses/LICENSE-2.0.
 # 
 # SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
 # 
 # Contributors:
 #     Stephan Wahlbrink <sw@wahlbrink.eu> - initial API and implementation
 #=============================================================================*/

package org.eclipse.statet.jcommons.text.core.util;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.statet.jcommons.lang.NonNullByDefault;
import org.eclipse.statet.jcommons.lang.Nullable;


@NonNullByDefault
public class HtmlUtils {
	
	
	public final static class Entity {
		
		private final String name;
		private final String string;
		private final char[] chars;
		
		
		public Entity(final String name, final String string) {
			this.name= name;
			this.string= string;
			this.chars= string.toCharArray();
		}
		
		
		public String getName() {
			return this.name;
		}
		
		public String getString() {
			return this.string;
		}
		
		char[] getChars() {
			return this.chars;
		}
		
		
		@Override
		public int hashCode() {
			return this.name.hashCode();
		}
		
		@Override
		public boolean equals(final @Nullable Object obj) {
			return ((this == obj)
					|| (obj instanceof Entity && this.name == ((Entity) obj).name) );
		}
		
		@Override
		public String toString() {
			return this.name;
		}
		
	}
	
	
	private static final Map<String, Entity> NAMED_ENTITIES;
	static {
		final Map<String, Entity> entities= new HashMap<>();
		add(entities, "Aacute", "\u00C1");
		add(entities, "aacute", "\u00E1");
		add(entities, "Abreve", "\u0102");
		add(entities, "abreve", "\u0103");
		add(entities, "ac", "\u223E");
		add(entities, "acd", "\u223F");
		add(entities, "acE", "\u223E\u0333");
		add(entities, "Acirc", "\u00C2");
		add(entities, "acirc", "\u00E2");
		add(entities, "acute", "\u00B4");
		add(entities, "Acy", "\u0410");
		add(entities, "acy", "\u0430");
		add(entities, "AElig", "\u00C6");
		add(entities, "aelig", "\u00E6");
		add(entities, "af", "\u2061");
		add(entities, "Afr", "\uD835\uDD04");
		add(entities, "afr", "\uD835\uDD1E");
		add(entities, "Agrave", "\u00C0");
		add(entities, "agrave", "\u00E0");
		add(entities, "alefsym", "\u2135");
		add(entities, "aleph", "\u2135");
		add(entities, "Alpha", "\u0391");
		add(entities, "alpha", "\u03B1");
		add(entities, "Amacr", "\u0100");
		add(entities, "amacr", "\u0101");
		add(entities, "amalg", "\u2A3F");
		add(entities, "AMP", "\u0026");
		add(entities, "amp", "\u0026");
		add(entities, "And", "\u2A53");
		add(entities, "and", "\u2227");
		add(entities, "andand", "\u2A55");
		add(entities, "andd", "\u2A5C");
		add(entities, "andslope", "\u2A58");
		add(entities, "andv", "\u2A5A");
		add(entities, "ang", "\u2220");
		add(entities, "ange", "\u29A4");
		add(entities, "angle", "\u2220");
		add(entities, "angmsd", "\u2221");
		add(entities, "angmsdaa", "\u29A8");
		add(entities, "angmsdab", "\u29A9");
		add(entities, "angmsdac", "\u29AA");
		add(entities, "angmsdad", "\u29AB");
		add(entities, "angmsdae", "\u29AC");
		add(entities, "angmsdaf", "\u29AD");
		add(entities, "angmsdag", "\u29AE");
		add(entities, "angmsdah", "\u29AF");
		add(entities, "angrt", "\u221F");
		add(entities, "angrtvb", "\u22BE");
		add(entities, "angrtvbd", "\u299D");
		add(entities, "angsph", "\u2222");
		add(entities, "angst", "\u00C5");
		add(entities, "angzarr", "\u237C");
		add(entities, "Aogon", "\u0104");
		add(entities, "aogon", "\u0105");
		add(entities, "Aopf", "\uD835\uDD38");
		add(entities, "aopf", "\uD835\uDD52");
		add(entities, "ap", "\u2248");
		add(entities, "apacir", "\u2A6F");
		add(entities, "apE", "\u2A70");
		add(entities, "ape", "\u224A");
		add(entities, "apid", "\u224B");
		add(entities, "apos", "\u0027");
		add(entities, "ApplyFunction", "\u2061");
		add(entities, "approx", "\u2248");
		add(entities, "approxeq", "\u224A");
		add(entities, "Aring", "\u00C5");
		add(entities, "aring", "\u00E5");
		add(entities, "Ascr", "\uD835\uDC9C");
		add(entities, "ascr", "\uD835\uDCB6");
		add(entities, "Assign", "\u2254");
		add(entities, "ast", "\u002A");
		add(entities, "asymp", "\u2248");
		add(entities, "asympeq", "\u224D");
		add(entities, "Atilde", "\u00C3");
		add(entities, "atilde", "\u00E3");
		add(entities, "Auml", "\u00C4");
		add(entities, "auml", "\u00E4");
		add(entities, "awconint", "\u2233");
		add(entities, "awint", "\u2A11");
		add(entities, "backcong", "\u224C");
		add(entities, "backepsilon", "\u03F6");
		add(entities, "backprime", "\u2035");
		add(entities, "backsim", "\u223D");
		add(entities, "backsimeq", "\u22CD");
		add(entities, "Backslash", "\u2216");
		add(entities, "Barv", "\u2AE7");
		add(entities, "barvee", "\u22BD");
		add(entities, "Barwed", "\u2306");
		add(entities, "barwed", "\u2305");
		add(entities, "barwedge", "\u2305");
		add(entities, "bbrk", "\u23B5");
		add(entities, "bbrktbrk", "\u23B6");
		add(entities, "bcong", "\u224C");
		add(entities, "Bcy", "\u0411");
		add(entities, "bcy", "\u0431");
		add(entities, "bdquo", "\u201E");
		add(entities, "becaus", "\u2235");
		add(entities, "Because", "\u2235");
		add(entities, "because", "\u2235");
		add(entities, "bemptyv", "\u29B0");
		add(entities, "bepsi", "\u03F6");
		add(entities, "bernou", "\u212C");
		add(entities, "Bernoullis", "\u212C");
		add(entities, "Beta", "\u0392");
		add(entities, "beta", "\u03B2");
		add(entities, "beth", "\u2136");
		add(entities, "between", "\u226C");
		add(entities, "Bfr", "\uD835\uDD05");
		add(entities, "bfr", "\uD835\uDD1F");
		add(entities, "bigcap", "\u22C2");
		add(entities, "bigcirc", "\u25EF");
		add(entities, "bigcup", "\u22C3");
		add(entities, "bigodot", "\u2A00");
		add(entities, "bigoplus", "\u2A01");
		add(entities, "bigotimes", "\u2A02");
		add(entities, "bigsqcup", "\u2A06");
		add(entities, "bigstar", "\u2605");
		add(entities, "bigtriangledown", "\u25BD");
		add(entities, "bigtriangleup", "\u25B3");
		add(entities, "biguplus", "\u2A04");
		add(entities, "bigvee", "\u22C1");
		add(entities, "bigwedge", "\u22C0");
		add(entities, "bkarow", "\u290D");
		add(entities, "blacklozenge", "\u29EB");
		add(entities, "blacksquare", "\u25AA");
		add(entities, "blacktriangle", "\u25B4");
		add(entities, "blacktriangledown", "\u25BE");
		add(entities, "blacktriangleleft", "\u25C2");
		add(entities, "blacktriangleright", "\u25B8");
		add(entities, "blank", "\u2423");
		add(entities, "blk12", "\u2592");
		add(entities, "blk14", "\u2591");
		add(entities, "blk34", "\u2593");
		add(entities, "block", "\u2588");
		add(entities, "bne", "\u003D\u20E5");
		add(entities, "bnequiv", "\u2261\u20E5");
		add(entities, "bNot", "\u2AED");
		add(entities, "bnot", "\u2310");
		add(entities, "Bopf", "\uD835\uDD39");
		add(entities, "bopf", "\uD835\uDD53");
		add(entities, "bot", "\u22A5");
		add(entities, "bottom", "\u22A5");
		add(entities, "bowtie", "\u22C8");
		add(entities, "boxbox", "\u29C9");
		add(entities, "boxDL", "\u2557");
		add(entities, "boxDl", "\u2556");
		add(entities, "boxdL", "\u2555");
		add(entities, "boxdl", "\u2510");
		add(entities, "boxDR", "\u2554");
		add(entities, "boxDr", "\u2553");
		add(entities, "boxdR", "\u2552");
		add(entities, "boxdr", "\u250C");
		add(entities, "boxH", "\u2550");
		add(entities, "boxh", "\u2500");
		add(entities, "boxHD", "\u2566");
		add(entities, "boxHd", "\u2564");
		add(entities, "boxhD", "\u2565");
		add(entities, "boxhd", "\u252C");
		add(entities, "boxHU", "\u2569");
		add(entities, "boxHu", "\u2567");
		add(entities, "boxhU", "\u2568");
		add(entities, "boxhu", "\u2534");
		add(entities, "boxminus", "\u229F");
		add(entities, "boxplus", "\u229E");
		add(entities, "boxtimes", "\u22A0");
		add(entities, "boxUL", "\u255D");
		add(entities, "boxUl", "\u255C");
		add(entities, "boxuL", "\u255B");
		add(entities, "boxul", "\u2518");
		add(entities, "boxUR", "\u255A");
		add(entities, "boxUr", "\u2559");
		add(entities, "boxuR", "\u2558");
		add(entities, "boxur", "\u2514");
		add(entities, "boxV", "\u2551");
		add(entities, "boxv", "\u2502");
		add(entities, "boxVH", "\u256C");
		add(entities, "boxVh", "\u256B");
		add(entities, "boxvH", "\u256A");
		add(entities, "boxvh", "\u253C");
		add(entities, "boxVL", "\u2563");
		add(entities, "boxVl", "\u2562");
		add(entities, "boxvL", "\u2561");
		add(entities, "boxvl", "\u2524");
		add(entities, "boxVR", "\u2560");
		add(entities, "boxVr", "\u255F");
		add(entities, "boxvR", "\u255E");
		add(entities, "boxvr", "\u251C");
		add(entities, "bprime", "\u2035");
		add(entities, "Breve", "\u02D8");
		add(entities, "breve", "\u02D8");
		add(entities, "brvbar", "\u00A6");
		add(entities, "Bscr", "\u212C");
		add(entities, "bscr", "\uD835\uDCB7");
		add(entities, "bsemi", "\u204F");
		add(entities, "bsim", "\u223D");
		add(entities, "bsime", "\u22CD");
		add(entities, "bsol", "\\"); // u005C
		add(entities, "bsolb", "\u29C5");
		add(entities, "bsolhsub", "\u27C8");
		add(entities, "bull", "\u2022");
		add(entities, "bullet", "\u2022");
		add(entities, "bump", "\u224E");
		add(entities, "bumpE", "\u2AAE");
		add(entities, "bumpe", "\u224F");
		add(entities, "Bumpeq", "\u224E");
		add(entities, "bumpeq", "\u224F");
		add(entities, "Cacute", "\u0106");
		add(entities, "cacute", "\u0107");
		add(entities, "Cap", "\u22D2");
		add(entities, "cap", "\u2229");
		add(entities, "capand", "\u2A44");
		add(entities, "capbrcup", "\u2A49");
		add(entities, "capcap", "\u2A4B");
		add(entities, "capcup", "\u2A47");
		add(entities, "capdot", "\u2A40");
		add(entities, "CapitalDifferentialD", "\u2145");
		add(entities, "caps", "\u2229\uFE00");
		add(entities, "caret", "\u2041");
		add(entities, "caron", "\u02C7");
		add(entities, "Cayleys", "\u212D");
		add(entities, "ccaps", "\u2A4D");
		add(entities, "Ccaron", "\u010C");
		add(entities, "ccaron", "\u010D");
		add(entities, "Ccedil", "\u00C7");
		add(entities, "ccedil", "\u00E7");
		add(entities, "Ccirc", "\u0108");
		add(entities, "ccirc", "\u0109");
		add(entities, "Cconint", "\u2230");
		add(entities, "ccups", "\u2A4C");
		add(entities, "ccupssm", "\u2A50");
		add(entities, "Cdot", "\u010A");
		add(entities, "cdot", "\u010B");
		add(entities, "cedil", "\u00B8");
		add(entities, "Cedilla", "\u00B8");
		add(entities, "cemptyv", "\u29B2");
		add(entities, "cent", "\u00A2");
		add(entities, "CenterDot", "\u00B7");
		add(entities, "centerdot", "\u00B7");
		add(entities, "Cfr", "\u212D");
		add(entities, "cfr", "\uD835\uDD20");
		add(entities, "CHcy", "\u0427");
		add(entities, "chcy", "\u0447");
		add(entities, "check", "\u2713");
		add(entities, "checkmark", "\u2713");
		add(entities, "Chi", "\u03A7");
		add(entities, "chi", "\u03C7");
		add(entities, "cir", "\u25CB");
		add(entities, "circ", "\u02C6");
		add(entities, "circeq", "\u2257");
		add(entities, "circlearrowleft", "\u21BA");
		add(entities, "circlearrowright", "\u21BB");
		add(entities, "circledast", "\u229B");
		add(entities, "circledcirc", "\u229A");
		add(entities, "circleddash", "\u229D");
		add(entities, "CircleDot", "\u2299");
		add(entities, "circledR", "\u00AE");
		add(entities, "circledS", "\u24C8");
		add(entities, "CircleMinus", "\u2296");
		add(entities, "CirclePlus", "\u2295");
		add(entities, "CircleTimes", "\u2297");
		add(entities, "cirE", "\u29C3");
		add(entities, "cire", "\u2257");
		add(entities, "cirfnint", "\u2A10");
		add(entities, "cirmid", "\u2AEF");
		add(entities, "cirscir", "\u29C2");
		add(entities, "ClockwiseContourIntegral", "\u2232");
		add(entities, "CloseCurlyDoubleQuote", "\u201D");
		add(entities, "CloseCurlyQuote", "\u2019");
		add(entities, "clubs", "\u2663");
		add(entities, "clubsuit", "\u2663");
		add(entities, "Colon", "\u2237");
		add(entities, "colon", "\u003A");
		add(entities, "Colone", "\u2A74");
		add(entities, "colone", "\u2254");
		add(entities, "coloneq", "\u2254");
		add(entities, "comma", "\u002C");
		add(entities, "commat", "\u0040");
		add(entities, "comp", "\u2201");
		add(entities, "compfn", "\u2218");
		add(entities, "complement", "\u2201");
		add(entities, "complexes", "\u2102");
		add(entities, "cong", "\u2245");
		add(entities, "congdot", "\u2A6D");
		add(entities, "Congruent", "\u2261");
		add(entities, "Conint", "\u222F");
		add(entities, "conint", "\u222E");
		add(entities, "ContourIntegral", "\u222E");
		add(entities, "Copf", "\u2102");
		add(entities, "copf", "\uD835\uDD54");
		add(entities, "coprod", "\u2210");
		add(entities, "Coproduct", "\u2210");
		add(entities, "COPY", "\u00A9");
		add(entities, "copy", "\u00A9");
		add(entities, "copysr", "\u2117");
		add(entities, "CounterClockwiseContourIntegral", "\u2233");
		add(entities, "crarr", "\u21B5");
		add(entities, "Cross", "\u2A2F");
		add(entities, "cross", "\u2717");
		add(entities, "Cscr", "\uD835\uDC9E");
		add(entities, "cscr", "\uD835\uDCB8");
		add(entities, "csub", "\u2ACF");
		add(entities, "csube", "\u2AD1");
		add(entities, "csup", "\u2AD0");
		add(entities, "csupe", "\u2AD2");
		add(entities, "ctdot", "\u22EF");
		add(entities, "cudarrl", "\u2938");
		add(entities, "cudarrr", "\u2935");
		add(entities, "cuepr", "\u22DE");
		add(entities, "cuesc", "\u22DF");
		add(entities, "cularr", "\u21B6");
		add(entities, "cularrp", "\u293D");
		add(entities, "Cup", "\u22D3");
		add(entities, "cup", "\u222A");
		add(entities, "cupbrcap", "\u2A48");
		add(entities, "CupCap", "\u224D");
		add(entities, "cupcap", "\u2A46");
		add(entities, "cupcup", "\u2A4A");
		add(entities, "cupdot", "\u228D");
		add(entities, "cupor", "\u2A45");
		add(entities, "cups", "\u222A\uFE00");
		add(entities, "curarr", "\u21B7");
		add(entities, "curarrm", "\u293C");
		add(entities, "curlyeqprec", "\u22DE");
		add(entities, "curlyeqsucc", "\u22DF");
		add(entities, "curlyvee", "\u22CE");
		add(entities, "curlywedge", "\u22CF");
		add(entities, "curren", "\u00A4");
		add(entities, "curvearrowleft", "\u21B6");
		add(entities, "curvearrowright", "\u21B7");
		add(entities, "cuvee", "\u22CE");
		add(entities, "cuwed", "\u22CF");
		add(entities, "cwconint", "\u2232");
		add(entities, "cwint", "\u2231");
		add(entities, "cylcty", "\u232D");
		add(entities, "Dagger", "\u2021");
		add(entities, "dagger", "\u2020");
		add(entities, "daleth", "\u2138");
		add(entities, "Darr", "\u21A1");
		add(entities, "dArr", "\u21D3");
		add(entities, "darr", "\u2193");
		add(entities, "dash", "\u2010");
		add(entities, "Dashv", "\u2AE4");
		add(entities, "dashv", "\u22A3");
		add(entities, "dbkarow", "\u290F");
		add(entities, "dblac", "\u02DD");
		add(entities, "Dcaron", "\u010E");
		add(entities, "dcaron", "\u010F");
		add(entities, "Dcy", "\u0414");
		add(entities, "dcy", "\u0434");
		add(entities, "DD", "\u2145");
		add(entities, "dd", "\u2146");
		add(entities, "ddagger", "\u2021");
		add(entities, "ddarr", "\u21CA");
		add(entities, "DDotrahd", "\u2911");
		add(entities, "ddotseq", "\u2A77");
		add(entities, "deg", "\u00B0");
		add(entities, "Del", "\u2207");
		add(entities, "Delta", "\u0394");
		add(entities, "delta", "\u03B4");
		add(entities, "demptyv", "\u29B1");
		add(entities, "dfisht", "\u297F");
		add(entities, "Dfr", "\uD835\uDD07");
		add(entities, "dfr", "\uD835\uDD21");
		add(entities, "dHar", "\u2965");
		add(entities, "dharl", "\u21C3");
		add(entities, "dharr", "\u21C2");
		add(entities, "DiacriticalAcute", "\u00B4");
		add(entities, "DiacriticalDot", "\u02D9");
		add(entities, "DiacriticalDoubleAcute", "\u02DD");
		add(entities, "DiacriticalGrave", "\u0060");
		add(entities, "DiacriticalTilde", "\u02DC");
		add(entities, "diam", "\u22C4");
		add(entities, "Diamond", "\u22C4");
		add(entities, "diamond", "\u22C4");
		add(entities, "diamondsuit", "\u2666");
		add(entities, "diams", "\u2666");
		add(entities, "die", "\u00A8");
		add(entities, "DifferentialD", "\u2146");
		add(entities, "digamma", "\u03DD");
		add(entities, "disin", "\u22F2");
		add(entities, "div", "\u00F7");
		add(entities, "divide", "\u00F7");
		add(entities, "divideontimes", "\u22C7");
		add(entities, "divonx", "\u22C7");
		add(entities, "DJcy", "\u0402");
		add(entities, "djcy", "\u0452");
		add(entities, "dlcorn", "\u231E");
		add(entities, "dlcrop", "\u230D");
		add(entities, "dollar", "\u0024");
		add(entities, "Dopf", "\uD835\uDD3B");
		add(entities, "dopf", "\uD835\uDD55");
		add(entities, "Dot", "\u00A8");
		add(entities, "dot", "\u02D9");
		add(entities, "DotDot", "\u20DC");
		add(entities, "doteq", "\u2250");
		add(entities, "doteqdot", "\u2251");
		add(entities, "DotEqual", "\u2250");
		add(entities, "dotminus", "\u2238");
		add(entities, "dotplus", "\u2214");
		add(entities, "dotsquare", "\u22A1");
		add(entities, "doublebarwedge", "\u2306");
		add(entities, "DoubleContourIntegral", "\u222F");
		add(entities, "DoubleDot", "\u00A8");
		add(entities, "DoubleDownArrow", "\u21D3");
		add(entities, "DoubleLeftArrow", "\u21D0");
		add(entities, "DoubleLeftRightArrow", "\u21D4");
		add(entities, "DoubleLeftTee", "\u2AE4");
		add(entities, "DoubleLongLeftArrow", "\u27F8");
		add(entities, "DoubleLongLeftRightArrow", "\u27FA");
		add(entities, "DoubleLongRightArrow", "\u27F9");
		add(entities, "DoubleRightArrow", "\u21D2");
		add(entities, "DoubleRightTee", "\u22A8");
		add(entities, "DoubleUpArrow", "\u21D1");
		add(entities, "DoubleUpDownArrow", "\u21D5");
		add(entities, "DoubleVerticalBar", "\u2225");
		add(entities, "DownArrow", "\u2193");
		add(entities, "Downarrow", "\u21D3");
		add(entities, "downarrow", "\u2193");
		add(entities, "DownArrowBar", "\u2913");
		add(entities, "DownArrowUpArrow", "\u21F5");
		add(entities, "DownBreve", "\u0311");
		add(entities, "downdownarrows", "\u21CA");
		add(entities, "downharpoonleft", "\u21C3");
		add(entities, "downharpoonright", "\u21C2");
		add(entities, "DownLeftRightVector", "\u2950");
		add(entities, "DownLeftTeeVector", "\u295E");
		add(entities, "DownLeftVector", "\u21BD");
		add(entities, "DownLeftVectorBar", "\u2956");
		add(entities, "DownRightTeeVector", "\u295F");
		add(entities, "DownRightVector", "\u21C1");
		add(entities, "DownRightVectorBar", "\u2957");
		add(entities, "DownTee", "\u22A4");
		add(entities, "DownTeeArrow", "\u21A7");
		add(entities, "drbkarow", "\u2910");
		add(entities, "drcorn", "\u231F");
		add(entities, "drcrop", "\u230C");
		add(entities, "Dscr", "\uD835\uDC9F");
		add(entities, "dscr", "\uD835\uDCB9");
		add(entities, "DScy", "\u0405");
		add(entities, "dscy", "\u0455");
		add(entities, "dsol", "\u29F6");
		add(entities, "Dstrok", "\u0110");
		add(entities, "dstrok", "\u0111");
		add(entities, "dtdot", "\u22F1");
		add(entities, "dtri", "\u25BF");
		add(entities, "dtrif", "\u25BE");
		add(entities, "duarr", "\u21F5");
		add(entities, "duhar", "\u296F");
		add(entities, "dwangle", "\u29A6");
		add(entities, "DZcy", "\u040F");
		add(entities, "dzcy", "\u045F");
		add(entities, "dzigrarr", "\u27FF");
		add(entities, "Eacute", "\u00C9");
		add(entities, "eacute", "\u00E9");
		add(entities, "easter", "\u2A6E");
		add(entities, "Ecaron", "\u011A");
		add(entities, "ecaron", "\u011B");
		add(entities, "ecir", "\u2256");
		add(entities, "Ecirc", "\u00CA");
		add(entities, "ecirc", "\u00EA");
		add(entities, "ecolon", "\u2255");
		add(entities, "Ecy", "\u042D");
		add(entities, "ecy", "\u044D");
		add(entities, "eDDot", "\u2A77");
		add(entities, "Edot", "\u0116");
		add(entities, "eDot", "\u2251");
		add(entities, "edot", "\u0117");
		add(entities, "ee", "\u2147");
		add(entities, "efDot", "\u2252");
		add(entities, "Efr", "\uD835\uDD08");
		add(entities, "efr", "\uD835\uDD22");
		add(entities, "eg", "\u2A9A");
		add(entities, "Egrave", "\u00C8");
		add(entities, "egrave", "\u00E8");
		add(entities, "egs", "\u2A96");
		add(entities, "egsdot", "\u2A98");
		add(entities, "el", "\u2A99");
		add(entities, "Element", "\u2208");
		add(entities, "elinters", "\u23E7");
		add(entities, "ell", "\u2113");
		add(entities, "els", "\u2A95");
		add(entities, "elsdot", "\u2A97");
		add(entities, "Emacr", "\u0112");
		add(entities, "emacr", "\u0113");
		add(entities, "empty", "\u2205");
		add(entities, "emptyset", "\u2205");
		add(entities, "EmptySmallSquare", "\u25FB");
		add(entities, "emptyv", "\u2205");
		add(entities, "EmptyVerySmallSquare", "\u25AB");
		add(entities, "emsp", "\u2003");
		add(entities, "emsp13", "\u2004");
		add(entities, "emsp14", "\u2005");
		add(entities, "ENG", "\u014A");
		add(entities, "eng", "\u014B");
		add(entities, "ensp", "\u2002");
		add(entities, "Eogon", "\u0118");
		add(entities, "eogon", "\u0119");
		add(entities, "Eopf", "\uD835\uDD3C");
		add(entities, "eopf", "\uD835\uDD56");
		add(entities, "epar", "\u22D5");
		add(entities, "eparsl", "\u29E3");
		add(entities, "eplus", "\u2A71");
		add(entities, "epsi", "\u03B5");
		add(entities, "Epsilon", "\u0395");
		add(entities, "epsilon", "\u03B5");
		add(entities, "epsiv", "\u03F5");
		add(entities, "eqcirc", "\u2256");
		add(entities, "eqcolon", "\u2255");
		add(entities, "eqsim", "\u2242");
		add(entities, "eqslantgtr", "\u2A96");
		add(entities, "eqslantless", "\u2A95");
		add(entities, "Equal", "\u2A75");
		add(entities, "equals", "\u003D");
		add(entities, "EqualTilde", "\u2242");
		add(entities, "equest", "\u225F");
		add(entities, "Equilibrium", "\u21CC");
		add(entities, "equiv", "\u2261");
		add(entities, "equivDD", "\u2A78");
		add(entities, "eqvparsl", "\u29E5");
		add(entities, "erarr", "\u2971");
		add(entities, "erDot", "\u2253");
		add(entities, "Escr", "\u2130");
		add(entities, "escr", "\u212F");
		add(entities, "esdot", "\u2250");
		add(entities, "Esim", "\u2A73");
		add(entities, "esim", "\u2242");
		add(entities, "Eta", "\u0397");
		add(entities, "eta", "\u03B7");
		add(entities, "ETH", "\u00D0");
		add(entities, "eth", "\u00F0");
		add(entities, "Euml", "\u00CB");
		add(entities, "euml", "\u00EB");
		add(entities, "euro", "\u20AC");
		add(entities, "excl", "\u0021");
		add(entities, "exist", "\u2203");
		add(entities, "Exists", "\u2203");
		add(entities, "expectation", "\u2130");
		add(entities, "ExponentialE", "\u2147");
		add(entities, "exponentiale", "\u2147");
		add(entities, "fallingdotseq", "\u2252");
		add(entities, "Fcy", "\u0424");
		add(entities, "fcy", "\u0444");
		add(entities, "female", "\u2640");
		add(entities, "ffilig", "\uFB03");
		add(entities, "fflig", "\uFB00");
		add(entities, "ffllig", "\uFB04");
		add(entities, "Ffr", "\uD835\uDD09");
		add(entities, "ffr", "\uD835\uDD23");
		add(entities, "filig", "\uFB01");
		add(entities, "FilledSmallSquare", "\u25FC");
		add(entities, "FilledVerySmallSquare", "\u25AA");
		add(entities, "fjlig", "\u0066\u006A");
		add(entities, "flat", "\u266D");
		add(entities, "fllig", "\uFB02");
		add(entities, "fltns", "\u25B1");
		add(entities, "fnof", "\u0192");
		add(entities, "Fopf", "\uD835\uDD3D");
		add(entities, "fopf", "\uD835\uDD57");
		add(entities, "ForAll", "\u2200");
		add(entities, "forall", "\u2200");
		add(entities, "fork", "\u22D4");
		add(entities, "forkv", "\u2AD9");
		add(entities, "Fouriertrf", "\u2131");
		add(entities, "fpartint", "\u2A0D");
		add(entities, "frac12", "\u00BD");
		add(entities, "frac13", "\u2153");
		add(entities, "frac14", "\u00BC");
		add(entities, "frac15", "\u2155");
		add(entities, "frac16", "\u2159");
		add(entities, "frac18", "\u215B");
		add(entities, "frac23", "\u2154");
		add(entities, "frac25", "\u2156");
		add(entities, "frac34", "\u00BE");
		add(entities, "frac35", "\u2157");
		add(entities, "frac38", "\u215C");
		add(entities, "frac45", "\u2158");
		add(entities, "frac56", "\u215A");
		add(entities, "frac58", "\u215D");
		add(entities, "frac78", "\u215E");
		add(entities, "frasl", "\u2044");
		add(entities, "frown", "\u2322");
		add(entities, "Fscr", "\u2131");
		add(entities, "fscr", "\uD835\uDCBB");
		add(entities, "gacute", "\u01F5");
		add(entities, "Gamma", "\u0393");
		add(entities, "gamma", "\u03B3");
		add(entities, "Gammad", "\u03DC");
		add(entities, "gammad", "\u03DD");
		add(entities, "gap", "\u2A86");
		add(entities, "Gbreve", "\u011E");
		add(entities, "gbreve", "\u011F");
		add(entities, "Gcedil", "\u0122");
		add(entities, "Gcirc", "\u011C");
		add(entities, "gcirc", "\u011D");
		add(entities, "Gcy", "\u0413");
		add(entities, "gcy", "\u0433");
		add(entities, "Gdot", "\u0120");
		add(entities, "gdot", "\u0121");
		add(entities, "gE", "\u2267");
		add(entities, "ge", "\u2265");
		add(entities, "gEl", "\u2A8C");
		add(entities, "gel", "\u22DB");
		add(entities, "geq", "\u2265");
		add(entities, "geqq", "\u2267");
		add(entities, "geqslant", "\u2A7E");
		add(entities, "ges", "\u2A7E");
		add(entities, "gescc", "\u2AA9");
		add(entities, "gesdot", "\u2A80");
		add(entities, "gesdoto", "\u2A82");
		add(entities, "gesdotol", "\u2A84");
		add(entities, "gesl", "\u22DB\uFE00");
		add(entities, "gesles", "\u2A94");
		add(entities, "Gfr", "\uD835\uDD0A");
		add(entities, "gfr", "\uD835\uDD24");
		add(entities, "Gg", "\u22D9");
		add(entities, "gg", "\u226B");
		add(entities, "ggg", "\u22D9");
		add(entities, "gimel", "\u2137");
		add(entities, "GJcy", "\u0403");
		add(entities, "gjcy", "\u0453");
		add(entities, "gl", "\u2277");
		add(entities, "gla", "\u2AA5");
		add(entities, "glE", "\u2A92");
		add(entities, "glj", "\u2AA4");
		add(entities, "gnap", "\u2A8A");
		add(entities, "gnapprox", "\u2A8A");
		add(entities, "gnE", "\u2269");
		add(entities, "gne", "\u2A88");
		add(entities, "gneq", "\u2A88");
		add(entities, "gneqq", "\u2269");
		add(entities, "gnsim", "\u22E7");
		add(entities, "Gopf", "\uD835\uDD3E");
		add(entities, "gopf", "\uD835\uDD58");
		add(entities, "grave", "\u0060");
		add(entities, "GreaterEqual", "\u2265");
		add(entities, "GreaterEqualLess", "\u22DB");
		add(entities, "GreaterFullEqual", "\u2267");
		add(entities, "GreaterGreater", "\u2AA2");
		add(entities, "GreaterLess", "\u2277");
		add(entities, "GreaterSlantEqual", "\u2A7E");
		add(entities, "GreaterTilde", "\u2273");
		add(entities, "Gscr", "\uD835\uDCA2");
		add(entities, "gscr", "\u210A");
		add(entities, "gsim", "\u2273");
		add(entities, "gsime", "\u2A8E");
		add(entities, "gsiml", "\u2A90");
		add(entities, "GT", "\u003E");
		add(entities, "Gt", "\u226B");
		add(entities, "gt", "\u003E");
		add(entities, "gtcc", "\u2AA7");
		add(entities, "gtcir", "\u2A7A");
		add(entities, "gtdot", "\u22D7");
		add(entities, "gtlPar", "\u2995");
		add(entities, "gtquest", "\u2A7C");
		add(entities, "gtrapprox", "\u2A86");
		add(entities, "gtrarr", "\u2978");
		add(entities, "gtrdot", "\u22D7");
		add(entities, "gtreqless", "\u22DB");
		add(entities, "gtreqqless", "\u2A8C");
		add(entities, "gtrless", "\u2277");
		add(entities, "gtrsim", "\u2273");
		add(entities, "gvertneqq", "\u2269\uFE00");
		add(entities, "gvnE", "\u2269\uFE00");
		add(entities, "Hacek", "\u02C7");
		add(entities, "hairsp", "\u200A");
		add(entities, "half", "\u00BD");
		add(entities, "hamilt", "\u210B");
		add(entities, "HARDcy", "\u042A");
		add(entities, "hardcy", "\u044A");
		add(entities, "hArr", "\u21D4");
		add(entities, "harr", "\u2194");
		add(entities, "harrcir", "\u2948");
		add(entities, "harrw", "\u21AD");
		add(entities, "Hat", "\u005E");
		add(entities, "hbar", "\u210F");
		add(entities, "Hcirc", "\u0124");
		add(entities, "hcirc", "\u0125");
		add(entities, "hearts", "\u2665");
		add(entities, "heartsuit", "\u2665");
		add(entities, "hellip", "\u2026");
		add(entities, "hercon", "\u22B9");
		add(entities, "Hfr", "\u210C");
		add(entities, "hfr", "\uD835\uDD25");
		add(entities, "HilbertSpace", "\u210B");
		add(entities, "hksearow", "\u2925");
		add(entities, "hkswarow", "\u2926");
		add(entities, "hoarr", "\u21FF");
		add(entities, "homtht", "\u223B");
		add(entities, "hookleftarrow", "\u21A9");
		add(entities, "hookrightarrow", "\u21AA");
		add(entities, "Hopf", "\u210D");
		add(entities, "hopf", "\uD835\uDD59");
		add(entities, "horbar", "\u2015");
		add(entities, "HorizontalLine", "\u2500");
		add(entities, "Hscr", "\u210B");
		add(entities, "hscr", "\uD835\uDCBD");
		add(entities, "hslash", "\u210F");
		add(entities, "Hstrok", "\u0126");
		add(entities, "hstrok", "\u0127");
		add(entities, "HumpDownHump", "\u224E");
		add(entities, "HumpEqual", "\u224F");
		add(entities, "hybull", "\u2043");
		add(entities, "hyphen", "\u2010");
		add(entities, "Iacute", "\u00CD");
		add(entities, "iacute", "\u00ED");
		add(entities, "ic", "\u2063");
		add(entities, "Icirc", "\u00CE");
		add(entities, "icirc", "\u00EE");
		add(entities, "Icy", "\u0418");
		add(entities, "icy", "\u0438");
		add(entities, "Idot", "\u0130");
		add(entities, "IEcy", "\u0415");
		add(entities, "iecy", "\u0435");
		add(entities, "iexcl", "\u00A1");
		add(entities, "iff", "\u21D4");
		add(entities, "Ifr", "\u2111");
		add(entities, "ifr", "\uD835\uDD26");
		add(entities, "Igrave", "\u00CC");
		add(entities, "igrave", "\u00EC");
		add(entities, "ii", "\u2148");
		add(entities, "iiiint", "\u2A0C");
		add(entities, "iiint", "\u222D");
		add(entities, "iinfin", "\u29DC");
		add(entities, "iiota", "\u2129");
		add(entities, "IJlig", "\u0132");
		add(entities, "ijlig", "\u0133");
		add(entities, "Im", "\u2111");
		add(entities, "Imacr", "\u012A");
		add(entities, "imacr", "\u012B");
		add(entities, "image", "\u2111");
		add(entities, "ImaginaryI", "\u2148");
		add(entities, "imagline", "\u2110");
		add(entities, "imagpart", "\u2111");
		add(entities, "imath", "\u0131");
		add(entities, "imof", "\u22B7");
		add(entities, "imped", "\u01B5");
		add(entities, "Implies", "\u21D2");
		add(entities, "in", "\u2208");
		add(entities, "incare", "\u2105");
		add(entities, "infin", "\u221E");
		add(entities, "infintie", "\u29DD");
		add(entities, "inodot", "\u0131");
		add(entities, "Int", "\u222C");
		add(entities, "int", "\u222B");
		add(entities, "intcal", "\u22BA");
		add(entities, "integers", "\u2124");
		add(entities, "Integral", "\u222B");
		add(entities, "intercal", "\u22BA");
		add(entities, "Intersection", "\u22C2");
		add(entities, "intlarhk", "\u2A17");
		add(entities, "intprod", "\u2A3C");
		add(entities, "InvisibleComma", "\u2063");
		add(entities, "InvisibleTimes", "\u2062");
		add(entities, "IOcy", "\u0401");
		add(entities, "iocy", "\u0451");
		add(entities, "Iogon", "\u012E");
		add(entities, "iogon", "\u012F");
		add(entities, "Iopf", "\uD835\uDD40");
		add(entities, "iopf", "\uD835\uDD5A");
		add(entities, "Iota", "\u0399");
		add(entities, "iota", "\u03B9");
		add(entities, "iprod", "\u2A3C");
		add(entities, "iquest", "\u00BF");
		add(entities, "Iscr", "\u2110");
		add(entities, "iscr", "\uD835\uDCBE");
		add(entities, "isin", "\u2208");
		add(entities, "isindot", "\u22F5");
		add(entities, "isinE", "\u22F9");
		add(entities, "isins", "\u22F4");
		add(entities, "isinsv", "\u22F3");
		add(entities, "isinv", "\u2208");
		add(entities, "it", "\u2062");
		add(entities, "Itilde", "\u0128");
		add(entities, "itilde", "\u0129");
		add(entities, "Iukcy", "\u0406");
		add(entities, "iukcy", "\u0456");
		add(entities, "Iuml", "\u00CF");
		add(entities, "iuml", "\u00EF");
		add(entities, "Jcirc", "\u0134");
		add(entities, "jcirc", "\u0135");
		add(entities, "Jcy", "\u0419");
		add(entities, "jcy", "\u0439");
		add(entities, "Jfr", "\uD835\uDD0D");
		add(entities, "jfr", "\uD835\uDD27");
		add(entities, "jmath", "\u0237");
		add(entities, "Jopf", "\uD835\uDD41");
		add(entities, "jopf", "\uD835\uDD5B");
		add(entities, "Jscr", "\uD835\uDCA5");
		add(entities, "jscr", "\uD835\uDCBF");
		add(entities, "Jsercy", "\u0408");
		add(entities, "jsercy", "\u0458");
		add(entities, "Jukcy", "\u0404");
		add(entities, "jukcy", "\u0454");
		add(entities, "Kappa", "\u039A");
		add(entities, "kappa", "\u03BA");
		add(entities, "kappav", "\u03F0");
		add(entities, "Kcedil", "\u0136");
		add(entities, "kcedil", "\u0137");
		add(entities, "Kcy", "\u041A");
		add(entities, "kcy", "\u043A");
		add(entities, "Kfr", "\uD835\uDD0E");
		add(entities, "kfr", "\uD835\uDD28");
		add(entities, "kgreen", "\u0138");
		add(entities, "KHcy", "\u0425");
		add(entities, "khcy", "\u0445");
		add(entities, "KJcy", "\u040C");
		add(entities, "kjcy", "\u045C");
		add(entities, "Kopf", "\uD835\uDD42");
		add(entities, "kopf", "\uD835\uDD5C");
		add(entities, "Kscr", "\uD835\uDCA6");
		add(entities, "kscr", "\uD835\uDCC0");
		add(entities, "lAarr", "\u21DA");
		add(entities, "Lacute", "\u0139");
		add(entities, "lacute", "\u013A");
		add(entities, "laemptyv", "\u29B4");
		add(entities, "lagran", "\u2112");
		add(entities, "Lambda", "\u039B");
		add(entities, "lambda", "\u03BB");
		add(entities, "Lang", "\u27EA");
		add(entities, "lang", "\u27E8");
		add(entities, "langd", "\u2991");
		add(entities, "langle", "\u27E8");
		add(entities, "lap", "\u2A85");
		add(entities, "Laplacetrf", "\u2112");
		add(entities, "laquo", "\u00AB");
		add(entities, "Larr", "\u219E");
		add(entities, "lArr", "\u21D0");
		add(entities, "larr", "\u2190");
		add(entities, "larrb", "\u21E4");
		add(entities, "larrbfs", "\u291F");
		add(entities, "larrfs", "\u291D");
		add(entities, "larrhk", "\u21A9");
		add(entities, "larrlp", "\u21AB");
		add(entities, "larrpl", "\u2939");
		add(entities, "larrsim", "\u2973");
		add(entities, "larrtl", "\u21A2");
		add(entities, "lat", "\u2AAB");
		add(entities, "lAtail", "\u291B");
		add(entities, "latail", "\u2919");
		add(entities, "late", "\u2AAD");
		add(entities, "lates", "\u2AAD\uFE00");
		add(entities, "lBarr", "\u290E");
		add(entities, "lbarr", "\u290C");
		add(entities, "lbbrk", "\u2772");
		add(entities, "lbrace", "\u007B");
		add(entities, "lbrack", "\u005B");
		add(entities, "lbrke", "\u298B");
		add(entities, "lbrksld", "\u298F");
		add(entities, "lbrkslu", "\u298D");
		add(entities, "Lcaron", "\u013D");
		add(entities, "lcaron", "\u013E");
		add(entities, "Lcedil", "\u013B");
		add(entities, "lcedil", "\u013C");
		add(entities, "lceil", "\u2308");
		add(entities, "lcub", "\u007B");
		add(entities, "Lcy", "\u041B");
		add(entities, "lcy", "\u043B");
		add(entities, "ldca", "\u2936");
		add(entities, "ldquo", "\u201C");
		add(entities, "ldquor", "\u201E");
		add(entities, "ldrdhar", "\u2967");
		add(entities, "ldrushar", "\u294B");
		add(entities, "ldsh", "\u21B2");
		add(entities, "lE", "\u2266");
		add(entities, "le", "\u2264");
		add(entities, "LeftAngleBracket", "\u27E8");
		add(entities, "LeftArrow", "\u2190");
		add(entities, "Leftarrow", "\u21D0");
		add(entities, "leftarrow", "\u2190");
		add(entities, "LeftArrowBar", "\u21E4");
		add(entities, "LeftArrowRightArrow", "\u21C6");
		add(entities, "leftarrowtail", "\u21A2");
		add(entities, "LeftCeiling", "\u2308");
		add(entities, "LeftDoubleBracket", "\u27E6");
		add(entities, "LeftDownTeeVector", "\u2961");
		add(entities, "LeftDownVector", "\u21C3");
		add(entities, "LeftDownVectorBar", "\u2959");
		add(entities, "LeftFloor", "\u230A");
		add(entities, "leftharpoondown", "\u21BD");
		add(entities, "leftharpoonup", "\u21BC");
		add(entities, "leftleftarrows", "\u21C7");
		add(entities, "LeftRightArrow", "\u2194");
		add(entities, "Leftrightarrow", "\u21D4");
		add(entities, "leftrightarrow", "\u2194");
		add(entities, "leftrightarrows", "\u21C6");
		add(entities, "leftrightharpoons", "\u21CB");
		add(entities, "leftrightsquigarrow", "\u21AD");
		add(entities, "LeftRightVector", "\u294E");
		add(entities, "LeftTee", "\u22A3");
		add(entities, "LeftTeeArrow", "\u21A4");
		add(entities, "LeftTeeVector", "\u295A");
		add(entities, "leftthreetimes", "\u22CB");
		add(entities, "LeftTriangle", "\u22B2");
		add(entities, "LeftTriangleBar", "\u29CF");
		add(entities, "LeftTriangleEqual", "\u22B4");
		add(entities, "LeftUpDownVector", "\u2951");
		add(entities, "LeftUpTeeVector", "\u2960");
		add(entities, "LeftUpVector", "\u21BF");
		add(entities, "LeftUpVectorBar", "\u2958");
		add(entities, "LeftVector", "\u21BC");
		add(entities, "LeftVectorBar", "\u2952");
		add(entities, "lEg", "\u2A8B");
		add(entities, "leg", "\u22DA");
		add(entities, "leq", "\u2264");
		add(entities, "leqq", "\u2266");
		add(entities, "leqslant", "\u2A7D");
		add(entities, "les", "\u2A7D");
		add(entities, "lescc", "\u2AA8");
		add(entities, "lesdot", "\u2A7F");
		add(entities, "lesdoto", "\u2A81");
		add(entities, "lesdotor", "\u2A83");
		add(entities, "lesg", "\u22DA\uFE00");
		add(entities, "lesges", "\u2A93");
		add(entities, "lessapprox", "\u2A85");
		add(entities, "lessdot", "\u22D6");
		add(entities, "lesseqgtr", "\u22DA");
		add(entities, "lesseqqgtr", "\u2A8B");
		add(entities, "LessEqualGreater", "\u22DA");
		add(entities, "LessFullEqual", "\u2266");
		add(entities, "LessGreater", "\u2276");
		add(entities, "lessgtr", "\u2276");
		add(entities, "LessLess", "\u2AA1");
		add(entities, "lesssim", "\u2272");
		add(entities, "LessSlantEqual", "\u2A7D");
		add(entities, "LessTilde", "\u2272");
		add(entities, "lfisht", "\u297C");
		add(entities, "lfloor", "\u230A");
		add(entities, "Lfr", "\uD835\uDD0F");
		add(entities, "lfr", "\uD835\uDD29");
		add(entities, "lg", "\u2276");
		add(entities, "lgE", "\u2A91");
		add(entities, "lHar", "\u2962");
		add(entities, "lhard", "\u21BD");
		add(entities, "lharu", "\u21BC");
		add(entities, "lharul", "\u296A");
		add(entities, "lhblk", "\u2584");
		add(entities, "LJcy", "\u0409");
		add(entities, "ljcy", "\u0459");
		add(entities, "Ll", "\u22D8");
		add(entities, "ll", "\u226A");
		add(entities, "llarr", "\u21C7");
		add(entities, "llcorner", "\u231E");
		add(entities, "Lleftarrow", "\u21DA");
		add(entities, "llhard", "\u296B");
		add(entities, "lltri", "\u25FA");
		add(entities, "Lmidot", "\u013F");
		add(entities, "lmidot", "\u0140");
		add(entities, "lmoust", "\u23B0");
		add(entities, "lmoustache", "\u23B0");
		add(entities, "lnap", "\u2A89");
		add(entities, "lnapprox", "\u2A89");
		add(entities, "lnE", "\u2268");
		add(entities, "lne", "\u2A87");
		add(entities, "lneq", "\u2A87");
		add(entities, "lneqq", "\u2268");
		add(entities, "lnsim", "\u22E6");
		add(entities, "loang", "\u27EC");
		add(entities, "loarr", "\u21FD");
		add(entities, "lobrk", "\u27E6");
		add(entities, "LongLeftArrow", "\u27F5");
		add(entities, "Longleftarrow", "\u27F8");
		add(entities, "longleftarrow", "\u27F5");
		add(entities, "LongLeftRightArrow", "\u27F7");
		add(entities, "Longleftrightarrow", "\u27FA");
		add(entities, "longleftrightarrow", "\u27F7");
		add(entities, "longmapsto", "\u27FC");
		add(entities, "LongRightArrow", "\u27F6");
		add(entities, "Longrightarrow", "\u27F9");
		add(entities, "longrightarrow", "\u27F6");
		add(entities, "looparrowleft", "\u21AB");
		add(entities, "looparrowright", "\u21AC");
		add(entities, "lopar", "\u2985");
		add(entities, "Lopf", "\uD835\uDD43");
		add(entities, "lopf", "\uD835\uDD5D");
		add(entities, "loplus", "\u2A2D");
		add(entities, "lotimes", "\u2A34");
		add(entities, "lowast", "\u2217");
		add(entities, "lowbar", "\u005F");
		add(entities, "LowerLeftArrow", "\u2199");
		add(entities, "LowerRightArrow", "\u2198");
		add(entities, "loz", "\u25CA");
		add(entities, "lozenge", "\u25CA");
		add(entities, "lozf", "\u29EB");
		add(entities, "lpar", "\u0028");
		add(entities, "lparlt", "\u2993");
		add(entities, "lrarr", "\u21C6");
		add(entities, "lrcorner", "\u231F");
		add(entities, "lrhar", "\u21CB");
		add(entities, "lrhard", "\u296D");
		add(entities, "lrm", "\u200E");
		add(entities, "lrtri", "\u22BF");
		add(entities, "lsaquo", "\u2039");
		add(entities, "Lscr", "\u2112");
		add(entities, "lscr", "\uD835\uDCC1");
		add(entities, "Lsh", "\u21B0");
		add(entities, "lsh", "\u21B0");
		add(entities, "lsim", "\u2272");
		add(entities, "lsime", "\u2A8D");
		add(entities, "lsimg", "\u2A8F");
		add(entities, "lsqb", "\u005B");
		add(entities, "lsquo", "\u2018");
		add(entities, "lsquor", "\u201A");
		add(entities, "Lstrok", "\u0141");
		add(entities, "lstrok", "\u0142");
		add(entities, "LT", "\u003C");
		add(entities, "Lt", "\u226A");
		add(entities, "lt", "\u003C");
		add(entities, "ltcc", "\u2AA6");
		add(entities, "ltcir", "\u2A79");
		add(entities, "ltdot", "\u22D6");
		add(entities, "lthree", "\u22CB");
		add(entities, "ltimes", "\u22C9");
		add(entities, "ltlarr", "\u2976");
		add(entities, "ltquest", "\u2A7B");
		add(entities, "ltri", "\u25C3");
		add(entities, "ltrie", "\u22B4");
		add(entities, "ltrif", "\u25C2");
		add(entities, "ltrPar", "\u2996");
		add(entities, "lurdshar", "\u294A");
		add(entities, "luruhar", "\u2966");
		add(entities, "lvertneqq", "\u2268\uFE00");
		add(entities, "lvnE", "\u2268\uFE00");
		add(entities, "macr", "\u00AF");
		add(entities, "male", "\u2642");
		add(entities, "malt", "\u2720");
		add(entities, "maltese", "\u2720");
		add(entities, "Map", "\u2905");
		add(entities, "map", "\u21A6");
		add(entities, "mapsto", "\u21A6");
		add(entities, "mapstodown", "\u21A7");
		add(entities, "mapstoleft", "\u21A4");
		add(entities, "mapstoup", "\u21A5");
		add(entities, "marker", "\u25AE");
		add(entities, "mcomma", "\u2A29");
		add(entities, "Mcy", "\u041C");
		add(entities, "mcy", "\u043C");
		add(entities, "mdash", "\u2014");
		add(entities, "mDDot", "\u223A");
		add(entities, "measuredangle", "\u2221");
		add(entities, "MediumSpace", "\u205F");
		add(entities, "Mellintrf", "\u2133");
		add(entities, "Mfr", "\uD835\uDD10");
		add(entities, "mfr", "\uD835\uDD2A");
		add(entities, "mho", "\u2127");
		add(entities, "micro", "\u00B5");
		add(entities, "mid", "\u2223");
		add(entities, "midast", "\u002A");
		add(entities, "midcir", "\u2AF0");
		add(entities, "middot", "\u00B7");
		add(entities, "minus", "\u2212");
		add(entities, "minusb", "\u229F");
		add(entities, "minusd", "\u2238");
		add(entities, "minusdu", "\u2A2A");
		add(entities, "MinusPlus", "\u2213");
		add(entities, "mlcp", "\u2ADB");
		add(entities, "mldr", "\u2026");
		add(entities, "mnplus", "\u2213");
		add(entities, "models", "\u22A7");
		add(entities, "Mopf", "\uD835\uDD44");
		add(entities, "mopf", "\uD835\uDD5E");
		add(entities, "mp", "\u2213");
		add(entities, "Mscr", "\u2133");
		add(entities, "mscr", "\uD835\uDCC2");
		add(entities, "mstpos", "\u223E");
		add(entities, "Mu", "\u039C");
		add(entities, "mu", "\u03BC");
		add(entities, "multimap", "\u22B8");
		add(entities, "mumap", "\u22B8");
		add(entities, "nabla", "\u2207");
		add(entities, "Nacute", "\u0143");
		add(entities, "nacute", "\u0144");
		add(entities, "nang", "\u2220\u20D2");
		add(entities, "nap", "\u2249");
		add(entities, "napE", "\u2A70\u0338");
		add(entities, "napid", "\u224B\u0338");
		add(entities, "napos", "\u0149");
		add(entities, "napprox", "\u2249");
		add(entities, "natur", "\u266E");
		add(entities, "natural", "\u266E");
		add(entities, "naturals", "\u2115");
		add(entities, "nbsp", "\u00A0");
		add(entities, "nbump", "\u224E\u0338");
		add(entities, "nbumpe", "\u224F\u0338");
		add(entities, "ncap", "\u2A43");
		add(entities, "Ncaron", "\u0147");
		add(entities, "ncaron", "\u0148");
		add(entities, "Ncedil", "\u0145");
		add(entities, "ncedil", "\u0146");
		add(entities, "ncong", "\u2247");
		add(entities, "ncongdot", "\u2A6D\u0338");
		add(entities, "ncup", "\u2A42");
		add(entities, "Ncy", "\u041D");
		add(entities, "ncy", "\u043D");
		add(entities, "ndash", "\u2013");
		add(entities, "ne", "\u2260");
		add(entities, "nearhk", "\u2924");
		add(entities, "neArr", "\u21D7");
		add(entities, "nearr", "\u2197");
		add(entities, "nearrow", "\u2197");
		add(entities, "nedot", "\u2250\u0338");
		add(entities, "NegativeMediumSpace", "\u200B");
		add(entities, "NegativeThickSpace", "\u200B");
		add(entities, "NegativeThinSpace", "\u200B");
		add(entities, "NegativeVeryThinSpace", "\u200B");
		add(entities, "nequiv", "\u2262");
		add(entities, "nesear", "\u2928");
		add(entities, "nesim", "\u2242\u0338");
		add(entities, "NestedGreaterGreater", "\u226B");
		add(entities, "NestedLessLess", "\u226A");
		add(entities, "NewLine", "\n"); // u000A
		add(entities, "nexist", "\u2204");
		add(entities, "nexists", "\u2204");
		add(entities, "Nfr", "\uD835\uDD11");
		add(entities, "nfr", "\uD835\uDD2B");
		add(entities, "ngE", "\u2267\u0338");
		add(entities, "nge", "\u2271");
		add(entities, "ngeq", "\u2271");
		add(entities, "ngeqq", "\u2267\u0338");
		add(entities, "ngeqslant", "\u2A7E\u0338");
		add(entities, "nges", "\u2A7E\u0338");
		add(entities, "nGg", "\u22D9\u0338");
		add(entities, "ngsim", "\u2275");
		add(entities, "nGt", "\u226B\u20D2");
		add(entities, "ngt", "\u226F");
		add(entities, "ngtr", "\u226F");
		add(entities, "nGtv", "\u226B\u0338");
		add(entities, "nhArr", "\u21CE");
		add(entities, "nharr", "\u21AE");
		add(entities, "nhpar", "\u2AF2");
		add(entities, "ni", "\u220B");
		add(entities, "nis", "\u22FC");
		add(entities, "nisd", "\u22FA");
		add(entities, "niv", "\u220B");
		add(entities, "NJcy", "\u040A");
		add(entities, "njcy", "\u045A");
		add(entities, "nlArr", "\u21CD");
		add(entities, "nlarr", "\u219A");
		add(entities, "nldr", "\u2025");
		add(entities, "nlE", "\u2266\u0338");
		add(entities, "nle", "\u2270");
		add(entities, "nLeftarrow", "\u21CD");
		add(entities, "nleftarrow", "\u219A");
		add(entities, "nLeftrightarrow", "\u21CE");
		add(entities, "nleftrightarrow", "\u21AE");
		add(entities, "nleq", "\u2270");
		add(entities, "nleqq", "\u2266\u0338");
		add(entities, "nleqslant", "\u2A7D\u0338");
		add(entities, "nles", "\u2A7D\u0338");
		add(entities, "nless", "\u226E");
		add(entities, "nLl", "\u22D8\u0338");
		add(entities, "nlsim", "\u2274");
		add(entities, "nLt", "\u226A\u20D2");
		add(entities, "nlt", "\u226E");
		add(entities, "nltri", "\u22EA");
		add(entities, "nltrie", "\u22EC");
		add(entities, "nLtv", "\u226A\u0338");
		add(entities, "nmid", "\u2224");
		add(entities, "NoBreak", "\u2060");
		add(entities, "NonBreakingSpace", "\u00A0");
		add(entities, "Nopf", "\u2115");
		add(entities, "nopf", "\uD835\uDD5F");
		add(entities, "Not", "\u2AEC");
		add(entities, "not", "\u00AC");
		add(entities, "NotCongruent", "\u2262");
		add(entities, "NotCupCap", "\u226D");
		add(entities, "NotDoubleVerticalBar", "\u2226");
		add(entities, "NotElement", "\u2209");
		add(entities, "NotEqual", "\u2260");
		add(entities, "NotEqualTilde", "\u2242\u0338");
		add(entities, "NotExists", "\u2204");
		add(entities, "NotGreater", "\u226F");
		add(entities, "NotGreaterEqual", "\u2271");
		add(entities, "NotGreaterFullEqual", "\u2267\u0338");
		add(entities, "NotGreaterGreater", "\u226B\u0338");
		add(entities, "NotGreaterLess", "\u2279");
		add(entities, "NotGreaterSlantEqual", "\u2A7E\u0338");
		add(entities, "NotGreaterTilde", "\u2275");
		add(entities, "NotHumpDownHump", "\u224E\u0338");
		add(entities, "NotHumpEqual", "\u224F\u0338");
		add(entities, "notin", "\u2209");
		add(entities, "notindot", "\u22F5\u0338");
		add(entities, "notinE", "\u22F9\u0338");
		add(entities, "notinva", "\u2209");
		add(entities, "notinvb", "\u22F7");
		add(entities, "notinvc", "\u22F6");
		add(entities, "NotLeftTriangle", "\u22EA");
		add(entities, "NotLeftTriangleBar", "\u29CF\u0338");
		add(entities, "NotLeftTriangleEqual", "\u22EC");
		add(entities, "NotLess", "\u226E");
		add(entities, "NotLessEqual", "\u2270");
		add(entities, "NotLessGreater", "\u2278");
		add(entities, "NotLessLess", "\u226A\u0338");
		add(entities, "NotLessSlantEqual", "\u2A7D\u0338");
		add(entities, "NotLessTilde", "\u2274");
		add(entities, "NotNestedGreaterGreater", "\u2AA2\u0338");
		add(entities, "NotNestedLessLess", "\u2AA1\u0338");
		add(entities, "notni", "\u220C");
		add(entities, "notniva", "\u220C");
		add(entities, "notnivb", "\u22FE");
		add(entities, "notnivc", "\u22FD");
		add(entities, "NotPrecedes", "\u2280");
		add(entities, "NotPrecedesEqual", "\u2AAF\u0338");
		add(entities, "NotPrecedesSlantEqual", "\u22E0");
		add(entities, "NotReverseElement", "\u220C");
		add(entities, "NotRightTriangle", "\u22EB");
		add(entities, "NotRightTriangleBar", "\u29D0\u0338");
		add(entities, "NotRightTriangleEqual", "\u22ED");
		add(entities, "NotSquareSubset", "\u228F\u0338");
		add(entities, "NotSquareSubsetEqual", "\u22E2");
		add(entities, "NotSquareSuperset", "\u2290\u0338");
		add(entities, "NotSquareSupersetEqual", "\u22E3");
		add(entities, "NotSubset", "\u2282\u20D2");
		add(entities, "NotSubsetEqual", "\u2288");
		add(entities, "NotSucceeds", "\u2281");
		add(entities, "NotSucceedsEqual", "\u2AB0\u0338");
		add(entities, "NotSucceedsSlantEqual", "\u22E1");
		add(entities, "NotSucceedsTilde", "\u227F\u0338");
		add(entities, "NotSuperset", "\u2283\u20D2");
		add(entities, "NotSupersetEqual", "\u2289");
		add(entities, "NotTilde", "\u2241");
		add(entities, "NotTildeEqual", "\u2244");
		add(entities, "NotTildeFullEqual", "\u2247");
		add(entities, "NotTildeTilde", "\u2249");
		add(entities, "NotVerticalBar", "\u2224");
		add(entities, "npar", "\u2226");
		add(entities, "nparallel", "\u2226");
		add(entities, "nparsl", "\u2AFD\u20E5");
		add(entities, "npart", "\u2202\u0338");
		add(entities, "npolint", "\u2A14");
		add(entities, "npr", "\u2280");
		add(entities, "nprcue", "\u22E0");
		add(entities, "npre", "\u2AAF\u0338");
		add(entities, "nprec", "\u2280");
		add(entities, "npreceq", "\u2AAF\u0338");
		add(entities, "nrArr", "\u21CF");
		add(entities, "nrarr", "\u219B");
		add(entities, "nrarrc", "\u2933\u0338");
		add(entities, "nrarrw", "\u219D\u0338");
		add(entities, "nRightarrow", "\u21CF");
		add(entities, "nrightarrow", "\u219B");
		add(entities, "nrtri", "\u22EB");
		add(entities, "nrtrie", "\u22ED");
		add(entities, "nsc", "\u2281");
		add(entities, "nsccue", "\u22E1");
		add(entities, "nsce", "\u2AB0\u0338");
		add(entities, "Nscr", "\uD835\uDCA9");
		add(entities, "nscr", "\uD835\uDCC3");
		add(entities, "nshortmid", "\u2224");
		add(entities, "nshortparallel", "\u2226");
		add(entities, "nsim", "\u2241");
		add(entities, "nsime", "\u2244");
		add(entities, "nsimeq", "\u2244");
		add(entities, "nsmid", "\u2224");
		add(entities, "nspar", "\u2226");
		add(entities, "nsqsube", "\u22E2");
		add(entities, "nsqsupe", "\u22E3");
		add(entities, "nsub", "\u2284");
		add(entities, "nsubE", "\u2AC5\u0338");
		add(entities, "nsube", "\u2288");
		add(entities, "nsubset", "\u2282\u20D2");
		add(entities, "nsubseteq", "\u2288");
		add(entities, "nsubseteqq", "\u2AC5\u0338");
		add(entities, "nsucc", "\u2281");
		add(entities, "nsucceq", "\u2AB0\u0338");
		add(entities, "nsup", "\u2285");
		add(entities, "nsupE", "\u2AC6\u0338");
		add(entities, "nsupe", "\u2289");
		add(entities, "nsupset", "\u2283\u20D2");
		add(entities, "nsupseteq", "\u2289");
		add(entities, "nsupseteqq", "\u2AC6\u0338");
		add(entities, "ntgl", "\u2279");
		add(entities, "Ntilde", "\u00D1");
		add(entities, "ntilde", "\u00F1");
		add(entities, "ntlg", "\u2278");
		add(entities, "ntriangleleft", "\u22EA");
		add(entities, "ntrianglelefteq", "\u22EC");
		add(entities, "ntriangleright", "\u22EB");
		add(entities, "ntrianglerighteq", "\u22ED");
		add(entities, "Nu", "\u039D");
		add(entities, "nu", "\u03BD");
		add(entities, "num", "\u0023");
		add(entities, "numero", "\u2116");
		add(entities, "numsp", "\u2007");
		add(entities, "nvap", "\u224D\u20D2");
		add(entities, "nVDash", "\u22AF");
		add(entities, "nVdash", "\u22AE");
		add(entities, "nvDash", "\u22AD");
		add(entities, "nvdash", "\u22AC");
		add(entities, "nvge", "\u2265\u20D2");
		add(entities, "nvgt", "\u003E\u20D2");
		add(entities, "nvHarr", "\u2904");
		add(entities, "nvinfin", "\u29DE");
		add(entities, "nvlArr", "\u2902");
		add(entities, "nvle", "\u2264\u20D2");
		add(entities, "nvlt", "\u003C\u20D2");
		add(entities, "nvltrie", "\u22B4\u20D2");
		add(entities, "nvrArr", "\u2903");
		add(entities, "nvrtrie", "\u22B5\u20D2");
		add(entities, "nvsim", "\u223C\u20D2");
		add(entities, "nwarhk", "\u2923");
		add(entities, "nwArr", "\u21D6");
		add(entities, "nwarr", "\u2196");
		add(entities, "nwarrow", "\u2196");
		add(entities, "nwnear", "\u2927");
		add(entities, "Oacute", "\u00D3");
		add(entities, "oacute", "\u00F3");
		add(entities, "oast", "\u229B");
		add(entities, "ocir", "\u229A");
		add(entities, "Ocirc", "\u00D4");
		add(entities, "ocirc", "\u00F4");
		add(entities, "Ocy", "\u041E");
		add(entities, "ocy", "\u043E");
		add(entities, "odash", "\u229D");
		add(entities, "Odblac", "\u0150");
		add(entities, "odblac", "\u0151");
		add(entities, "odiv", "\u2A38");
		add(entities, "odot", "\u2299");
		add(entities, "odsold", "\u29BC");
		add(entities, "OElig", "\u0152");
		add(entities, "oelig", "\u0153");
		add(entities, "ofcir", "\u29BF");
		add(entities, "Ofr", "\uD835\uDD12");
		add(entities, "ofr", "\uD835\uDD2C");
		add(entities, "ogon", "\u02DB");
		add(entities, "Ograve", "\u00D2");
		add(entities, "ograve", "\u00F2");
		add(entities, "ogt", "\u29C1");
		add(entities, "ohbar", "\u29B5");
		add(entities, "ohm", "\u03A9");
		add(entities, "oint", "\u222E");
		add(entities, "olarr", "\u21BA");
		add(entities, "olcir", "\u29BE");
		add(entities, "olcross", "\u29BB");
		add(entities, "oline", "\u203E");
		add(entities, "olt", "\u29C0");
		add(entities, "Omacr", "\u014C");
		add(entities, "omacr", "\u014D");
		add(entities, "Omega", "\u03A9");
		add(entities, "omega", "\u03C9");
		add(entities, "Omicron", "\u039F");
		add(entities, "omicron", "\u03BF");
		add(entities, "omid", "\u29B6");
		add(entities, "ominus", "\u2296");
		add(entities, "Oopf", "\uD835\uDD46");
		add(entities, "oopf", "\uD835\uDD60");
		add(entities, "opar", "\u29B7");
		add(entities, "OpenCurlyDoubleQuote", "\u201C");
		add(entities, "OpenCurlyQuote", "\u2018");
		add(entities, "operp", "\u29B9");
		add(entities, "oplus", "\u2295");
		add(entities, "Or", "\u2A54");
		add(entities, "or", "\u2228");
		add(entities, "orarr", "\u21BB");
		add(entities, "ord", "\u2A5D");
		add(entities, "order", "\u2134");
		add(entities, "orderof", "\u2134");
		add(entities, "ordf", "\u00AA");
		add(entities, "ordm", "\u00BA");
		add(entities, "origof", "\u22B6");
		add(entities, "oror", "\u2A56");
		add(entities, "orslope", "\u2A57");
		add(entities, "orv", "\u2A5B");
		add(entities, "oS", "\u24C8");
		add(entities, "Oscr", "\uD835\uDCAA");
		add(entities, "oscr", "\u2134");
		add(entities, "Oslash", "\u00D8");
		add(entities, "oslash", "\u00F8");
		add(entities, "osol", "\u2298");
		add(entities, "Otilde", "\u00D5");
		add(entities, "otilde", "\u00F5");
		add(entities, "Otimes", "\u2A37");
		add(entities, "otimes", "\u2297");
		add(entities, "otimesas", "\u2A36");
		add(entities, "Ouml", "\u00D6");
		add(entities, "ouml", "\u00F6");
		add(entities, "ovbar", "\u233D");
		add(entities, "OverBar", "\u203E");
		add(entities, "OverBrace", "\u23DE");
		add(entities, "OverBracket", "\u23B4");
		add(entities, "OverParenthesis", "\u23DC");
		add(entities, "par", "\u2225");
		add(entities, "para", "\u00B6");
		add(entities, "parallel", "\u2225");
		add(entities, "parsim", "\u2AF3");
		add(entities, "parsl", "\u2AFD");
		add(entities, "part", "\u2202");
		add(entities, "PartialD", "\u2202");
		add(entities, "Pcy", "\u041F");
		add(entities, "pcy", "\u043F");
		add(entities, "percnt", "\u0025");
		add(entities, "period", "\u002E");
		add(entities, "permil", "\u2030");
		add(entities, "perp", "\u22A5");
		add(entities, "pertenk", "\u2031");
		add(entities, "Pfr", "\uD835\uDD13");
		add(entities, "pfr", "\uD835\uDD2D");
		add(entities, "Phi", "\u03A6");
		add(entities, "phi", "\u03C6");
		add(entities, "phiv", "\u03D5");
		add(entities, "phmmat", "\u2133");
		add(entities, "phone", "\u260E");
		add(entities, "Pi", "\u03A0");
		add(entities, "pi", "\u03C0");
		add(entities, "pitchfork", "\u22D4");
		add(entities, "piv", "\u03D6");
		add(entities, "planck", "\u210F");
		add(entities, "planckh", "\u210E");
		add(entities, "plankv", "\u210F");
		add(entities, "plus", "\u002B");
		add(entities, "plusacir", "\u2A23");
		add(entities, "plusb", "\u229E");
		add(entities, "pluscir", "\u2A22");
		add(entities, "plusdo", "\u2214");
		add(entities, "plusdu", "\u2A25");
		add(entities, "pluse", "\u2A72");
		add(entities, "PlusMinus", "\u00B1");
		add(entities, "plusmn", "\u00B1");
		add(entities, "plussim", "\u2A26");
		add(entities, "plustwo", "\u2A27");
		add(entities, "pm", "\u00B1");
		add(entities, "Poincareplane", "\u210C");
		add(entities, "pointint", "\u2A15");
		add(entities, "Popf", "\u2119");
		add(entities, "popf", "\uD835\uDD61");
		add(entities, "pound", "\u00A3");
		add(entities, "Pr", "\u2ABB");
		add(entities, "pr", "\u227A");
		add(entities, "prap", "\u2AB7");
		add(entities, "prcue", "\u227C");
		add(entities, "prE", "\u2AB3");
		add(entities, "pre", "\u2AAF");
		add(entities, "prec", "\u227A");
		add(entities, "precapprox", "\u2AB7");
		add(entities, "preccurlyeq", "\u227C");
		add(entities, "Precedes", "\u227A");
		add(entities, "PrecedesEqual", "\u2AAF");
		add(entities, "PrecedesSlantEqual", "\u227C");
		add(entities, "PrecedesTilde", "\u227E");
		add(entities, "preceq", "\u2AAF");
		add(entities, "precnapprox", "\u2AB9");
		add(entities, "precneqq", "\u2AB5");
		add(entities, "precnsim", "\u22E8");
		add(entities, "precsim", "\u227E");
		add(entities, "Prime", "\u2033");
		add(entities, "prime", "\u2032");
		add(entities, "primes", "\u2119");
		add(entities, "prnap", "\u2AB9");
		add(entities, "prnE", "\u2AB5");
		add(entities, "prnsim", "\u22E8");
		add(entities, "prod", "\u220F");
		add(entities, "Product", "\u220F");
		add(entities, "profalar", "\u232E");
		add(entities, "profline", "\u2312");
		add(entities, "profsurf", "\u2313");
		add(entities, "prop", "\u221D");
		add(entities, "Proportion", "\u2237");
		add(entities, "Proportional", "\u221D");
		add(entities, "propto", "\u221D");
		add(entities, "prsim", "\u227E");
		add(entities, "prurel", "\u22B0");
		add(entities, "Pscr", "\uD835\uDCAB");
		add(entities, "pscr", "\uD835\uDCC5");
		add(entities, "Psi", "\u03A8");
		add(entities, "psi", "\u03C8");
		add(entities, "puncsp", "\u2008");
		add(entities, "Qfr", "\uD835\uDD14");
		add(entities, "qfr", "\uD835\uDD2E");
		add(entities, "qint", "\u2A0C");
		add(entities, "Qopf", "\u211A");
		add(entities, "qopf", "\uD835\uDD62");
		add(entities, "qprime", "\u2057");
		add(entities, "Qscr", "\uD835\uDCAC");
		add(entities, "qscr", "\uD835\uDCC6");
		add(entities, "quaternions", "\u210D");
		add(entities, "quatint", "\u2A16");
		add(entities, "quest", "\u003F");
		add(entities, "questeq", "\u225F");
		add(entities, "QUOT", "\""); // u0022
		add(entities, "quot", "\""); // u0022
		add(entities, "rAarr", "\u21DB");
		add(entities, "race", "\u223D\u0331");
		add(entities, "Racute", "\u0154");
		add(entities, "racute", "\u0155");
		add(entities, "radic", "\u221A");
		add(entities, "raemptyv", "\u29B3");
		add(entities, "Rang", "\u27EB");
		add(entities, "rang", "\u27E9");
		add(entities, "rangd", "\u2992");
		add(entities, "range", "\u29A5");
		add(entities, "rangle", "\u27E9");
		add(entities, "raquo", "\u00BB");
		add(entities, "Rarr", "\u21A0");
		add(entities, "rArr", "\u21D2");
		add(entities, "rarr", "\u2192");
		add(entities, "rarrap", "\u2975");
		add(entities, "rarrb", "\u21E5");
		add(entities, "rarrbfs", "\u2920");
		add(entities, "rarrc", "\u2933");
		add(entities, "rarrfs", "\u291E");
		add(entities, "rarrhk", "\u21AA");
		add(entities, "rarrlp", "\u21AC");
		add(entities, "rarrpl", "\u2945");
		add(entities, "rarrsim", "\u2974");
		add(entities, "Rarrtl", "\u2916");
		add(entities, "rarrtl", "\u21A3");
		add(entities, "rarrw", "\u219D");
		add(entities, "rAtail", "\u291C");
		add(entities, "ratail", "\u291A");
		add(entities, "ratio", "\u2236");
		add(entities, "rationals", "\u211A");
		add(entities, "RBarr", "\u2910");
		add(entities, "rBarr", "\u290F");
		add(entities, "rbarr", "\u290D");
		add(entities, "rbbrk", "\u2773");
		add(entities, "rbrace", "\u007D");
		add(entities, "rbrack", "\u005D");
		add(entities, "rbrke", "\u298C");
		add(entities, "rbrksld", "\u298E");
		add(entities, "rbrkslu", "\u2990");
		add(entities, "Rcaron", "\u0158");
		add(entities, "rcaron", "\u0159");
		add(entities, "Rcedil", "\u0156");
		add(entities, "rcedil", "\u0157");
		add(entities, "rceil", "\u2309");
		add(entities, "rcub", "\u007D");
		add(entities, "Rcy", "\u0420");
		add(entities, "rcy", "\u0440");
		add(entities, "rdca", "\u2937");
		add(entities, "rdldhar", "\u2969");
		add(entities, "rdquo", "\u201D");
		add(entities, "rdquor", "\u201D");
		add(entities, "rdsh", "\u21B3");
		add(entities, "Re", "\u211C");
		add(entities, "real", "\u211C");
		add(entities, "realine", "\u211B");
		add(entities, "realpart", "\u211C");
		add(entities, "reals", "\u211D");
		add(entities, "rect", "\u25AD");
		add(entities, "REG", "\u00AE");
		add(entities, "reg", "\u00AE");
		add(entities, "ReverseElement", "\u220B");
		add(entities, "ReverseEquilibrium", "\u21CB");
		add(entities, "ReverseUpEquilibrium", "\u296F");
		add(entities, "rfisht", "\u297D");
		add(entities, "rfloor", "\u230B");
		add(entities, "Rfr", "\u211C");
		add(entities, "rfr", "\uD835\uDD2F");
		add(entities, "rHar", "\u2964");
		add(entities, "rhard", "\u21C1");
		add(entities, "rharu", "\u21C0");
		add(entities, "rharul", "\u296C");
		add(entities, "Rho", "\u03A1");
		add(entities, "rho", "\u03C1");
		add(entities, "rhov", "\u03F1");
		add(entities, "RightAngleBracket", "\u27E9");
		add(entities, "RightArrow", "\u2192");
		add(entities, "Rightarrow", "\u21D2");
		add(entities, "rightarrow", "\u2192");
		add(entities, "RightArrowBar", "\u21E5");
		add(entities, "RightArrowLeftArrow", "\u21C4");
		add(entities, "rightarrowtail", "\u21A3");
		add(entities, "RightCeiling", "\u2309");
		add(entities, "RightDoubleBracket", "\u27E7");
		add(entities, "RightDownTeeVector", "\u295D");
		add(entities, "RightDownVector", "\u21C2");
		add(entities, "RightDownVectorBar", "\u2955");
		add(entities, "RightFloor", "\u230B");
		add(entities, "rightharpoondown", "\u21C1");
		add(entities, "rightharpoonup", "\u21C0");
		add(entities, "rightleftarrows", "\u21C4");
		add(entities, "rightleftharpoons", "\u21CC");
		add(entities, "rightrightarrows", "\u21C9");
		add(entities, "rightsquigarrow", "\u219D");
		add(entities, "RightTee", "\u22A2");
		add(entities, "RightTeeArrow", "\u21A6");
		add(entities, "RightTeeVector", "\u295B");
		add(entities, "rightthreetimes", "\u22CC");
		add(entities, "RightTriangle", "\u22B3");
		add(entities, "RightTriangleBar", "\u29D0");
		add(entities, "RightTriangleEqual", "\u22B5");
		add(entities, "RightUpDownVector", "\u294F");
		add(entities, "RightUpTeeVector", "\u295C");
		add(entities, "RightUpVector", "\u21BE");
		add(entities, "RightUpVectorBar", "\u2954");
		add(entities, "RightVector", "\u21C0");
		add(entities, "RightVectorBar", "\u2953");
		add(entities, "ring", "\u02DA");
		add(entities, "risingdotseq", "\u2253");
		add(entities, "rlarr", "\u21C4");
		add(entities, "rlhar", "\u21CC");
		add(entities, "rlm", "\u200F");
		add(entities, "rmoust", "\u23B1");
		add(entities, "rmoustache", "\u23B1");
		add(entities, "rnmid", "\u2AEE");
		add(entities, "roang", "\u27ED");
		add(entities, "roarr", "\u21FE");
		add(entities, "robrk", "\u27E7");
		add(entities, "ropar", "\u2986");
		add(entities, "Ropf", "\u211D");
		add(entities, "ropf", "\uD835\uDD63");
		add(entities, "roplus", "\u2A2E");
		add(entities, "rotimes", "\u2A35");
		add(entities, "RoundImplies", "\u2970");
		add(entities, "rpar", "\u0029");
		add(entities, "rpargt", "\u2994");
		add(entities, "rppolint", "\u2A12");
		add(entities, "rrarr", "\u21C9");
		add(entities, "Rrightarrow", "\u21DB");
		add(entities, "rsaquo", "\u203A");
		add(entities, "Rscr", "\u211B");
		add(entities, "rscr", "\uD835\uDCC7");
		add(entities, "Rsh", "\u21B1");
		add(entities, "rsh", "\u21B1");
		add(entities, "rsqb", "\u005D");
		add(entities, "rsquo", "\u2019");
		add(entities, "rsquor", "\u2019");
		add(entities, "rthree", "\u22CC");
		add(entities, "rtimes", "\u22CA");
		add(entities, "rtri", "\u25B9");
		add(entities, "rtrie", "\u22B5");
		add(entities, "rtrif", "\u25B8");
		add(entities, "rtriltri", "\u29CE");
		add(entities, "RuleDelayed", "\u29F4");
		add(entities, "ruluhar", "\u2968");
		add(entities, "rx", "\u211E");
		add(entities, "Sacute", "\u015A");
		add(entities, "sacute", "\u015B");
		add(entities, "sbquo", "\u201A");
		add(entities, "Sc", "\u2ABC");
		add(entities, "sc", "\u227B");
		add(entities, "scap", "\u2AB8");
		add(entities, "Scaron", "\u0160");
		add(entities, "scaron", "\u0161");
		add(entities, "sccue", "\u227D");
		add(entities, "scE", "\u2AB4");
		add(entities, "sce", "\u2AB0");
		add(entities, "Scedil", "\u015E");
		add(entities, "scedil", "\u015F");
		add(entities, "Scirc", "\u015C");
		add(entities, "scirc", "\u015D");
		add(entities, "scnap", "\u2ABA");
		add(entities, "scnE", "\u2AB6");
		add(entities, "scnsim", "\u22E9");
		add(entities, "scpolint", "\u2A13");
		add(entities, "scsim", "\u227F");
		add(entities, "Scy", "\u0421");
		add(entities, "scy", "\u0441");
		add(entities, "sdot", "\u22C5");
		add(entities, "sdotb", "\u22A1");
		add(entities, "sdote", "\u2A66");
		add(entities, "searhk", "\u2925");
		add(entities, "seArr", "\u21D8");
		add(entities, "searr", "\u2198");
		add(entities, "searrow", "\u2198");
		add(entities, "sect", "\u00A7");
		add(entities, "semi", "\u003B");
		add(entities, "seswar", "\u2929");
		add(entities, "setminus", "\u2216");
		add(entities, "setmn", "\u2216");
		add(entities, "sext", "\u2736");
		add(entities, "Sfr", "\uD835\uDD16");
		add(entities, "sfr", "\uD835\uDD30");
		add(entities, "sfrown", "\u2322");
		add(entities, "sharp", "\u266F");
		add(entities, "SHCHcy", "\u0429");
		add(entities, "shchcy", "\u0449");
		add(entities, "SHcy", "\u0428");
		add(entities, "shcy", "\u0448");
		add(entities, "ShortDownArrow", "\u2193");
		add(entities, "ShortLeftArrow", "\u2190");
		add(entities, "shortmid", "\u2223");
		add(entities, "shortparallel", "\u2225");
		add(entities, "ShortRightArrow", "\u2192");
		add(entities, "ShortUpArrow", "\u2191");
		add(entities, "shy", "\u00AD");
		add(entities, "Sigma", "\u03A3");
		add(entities, "sigma", "\u03C3");
		add(entities, "sigmaf", "\u03C2");
		add(entities, "sigmav", "\u03C2");
		add(entities, "sim", "\u223C");
		add(entities, "simdot", "\u2A6A");
		add(entities, "sime", "\u2243");
		add(entities, "simeq", "\u2243");
		add(entities, "simg", "\u2A9E");
		add(entities, "simgE", "\u2AA0");
		add(entities, "siml", "\u2A9D");
		add(entities, "simlE", "\u2A9F");
		add(entities, "simne", "\u2246");
		add(entities, "simplus", "\u2A24");
		add(entities, "simrarr", "\u2972");
		add(entities, "slarr", "\u2190");
		add(entities, "SmallCircle", "\u2218");
		add(entities, "smallsetminus", "\u2216");
		add(entities, "smashp", "\u2A33");
		add(entities, "smeparsl", "\u29E4");
		add(entities, "smid", "\u2223");
		add(entities, "smile", "\u2323");
		add(entities, "smt", "\u2AAA");
		add(entities, "smte", "\u2AAC");
		add(entities, "smtes", "\u2AAC\uFE00");
		add(entities, "SOFTcy", "\u042C");
		add(entities, "softcy", "\u044C");
		add(entities, "sol", "\u002F");
		add(entities, "solb", "\u29C4");
		add(entities, "solbar", "\u233F");
		add(entities, "Sopf", "\uD835\uDD4A");
		add(entities, "sopf", "\uD835\uDD64");
		add(entities, "spades", "\u2660");
		add(entities, "spadesuit", "\u2660");
		add(entities, "spar", "\u2225");
		add(entities, "sqcap", "\u2293");
		add(entities, "sqcaps", "\u2293\uFE00");
		add(entities, "sqcup", "\u2294");
		add(entities, "sqcups", "\u2294\uFE00");
		add(entities, "Sqrt", "\u221A");
		add(entities, "sqsub", "\u228F");
		add(entities, "sqsube", "\u2291");
		add(entities, "sqsubset", "\u228F");
		add(entities, "sqsubseteq", "\u2291");
		add(entities, "sqsup", "\u2290");
		add(entities, "sqsupe", "\u2292");
		add(entities, "sqsupset", "\u2290");
		add(entities, "sqsupseteq", "\u2292");
		add(entities, "squ", "\u25A1");
		add(entities, "Square", "\u25A1");
		add(entities, "square", "\u25A1");
		add(entities, "SquareIntersection", "\u2293");
		add(entities, "SquareSubset", "\u228F");
		add(entities, "SquareSubsetEqual", "\u2291");
		add(entities, "SquareSuperset", "\u2290");
		add(entities, "SquareSupersetEqual", "\u2292");
		add(entities, "SquareUnion", "\u2294");
		add(entities, "squarf", "\u25AA");
		add(entities, "squf", "\u25AA");
		add(entities, "srarr", "\u2192");
		add(entities, "Sscr", "\uD835\uDCAE");
		add(entities, "sscr", "\uD835\uDCC8");
		add(entities, "ssetmn", "\u2216");
		add(entities, "ssmile", "\u2323");
		add(entities, "sstarf", "\u22C6");
		add(entities, "Star", "\u22C6");
		add(entities, "star", "\u2606");
		add(entities, "starf", "\u2605");
		add(entities, "straightepsilon", "\u03F5");
		add(entities, "straightphi", "\u03D5");
		add(entities, "strns", "\u00AF");
		add(entities, "Sub", "\u22D0");
		add(entities, "sub", "\u2282");
		add(entities, "subdot", "\u2ABD");
		add(entities, "subE", "\u2AC5");
		add(entities, "sube", "\u2286");
		add(entities, "subedot", "\u2AC3");
		add(entities, "submult", "\u2AC1");
		add(entities, "subnE", "\u2ACB");
		add(entities, "subne", "\u228A");
		add(entities, "subplus", "\u2ABF");
		add(entities, "subrarr", "\u2979");
		add(entities, "Subset", "\u22D0");
		add(entities, "subset", "\u2282");
		add(entities, "subseteq", "\u2286");
		add(entities, "subseteqq", "\u2AC5");
		add(entities, "SubsetEqual", "\u2286");
		add(entities, "subsetneq", "\u228A");
		add(entities, "subsetneqq", "\u2ACB");
		add(entities, "subsim", "\u2AC7");
		add(entities, "subsub", "\u2AD5");
		add(entities, "subsup", "\u2AD3");
		add(entities, "succ", "\u227B");
		add(entities, "succapprox", "\u2AB8");
		add(entities, "succcurlyeq", "\u227D");
		add(entities, "Succeeds", "\u227B");
		add(entities, "SucceedsEqual", "\u2AB0");
		add(entities, "SucceedsSlantEqual", "\u227D");
		add(entities, "SucceedsTilde", "\u227F");
		add(entities, "succeq", "\u2AB0");
		add(entities, "succnapprox", "\u2ABA");
		add(entities, "succneqq", "\u2AB6");
		add(entities, "succnsim", "\u22E9");
		add(entities, "succsim", "\u227F");
		add(entities, "SuchThat", "\u220B");
		add(entities, "Sum", "\u2211");
		add(entities, "sum", "\u2211");
		add(entities, "sung", "\u266A");
		add(entities, "Sup", "\u22D1");
		add(entities, "sup", "\u2283");
		add(entities, "sup1", "\u00B9");
		add(entities, "sup2", "\u00B2");
		add(entities, "sup3", "\u00B3");
		add(entities, "supdot", "\u2ABE");
		add(entities, "supdsub", "\u2AD8");
		add(entities, "supE", "\u2AC6");
		add(entities, "supe", "\u2287");
		add(entities, "supedot", "\u2AC4");
		add(entities, "Superset", "\u2283");
		add(entities, "SupersetEqual", "\u2287");
		add(entities, "suphsol", "\u27C9");
		add(entities, "suphsub", "\u2AD7");
		add(entities, "suplarr", "\u297B");
		add(entities, "supmult", "\u2AC2");
		add(entities, "supnE", "\u2ACC");
		add(entities, "supne", "\u228B");
		add(entities, "supplus", "\u2AC0");
		add(entities, "Supset", "\u22D1");
		add(entities, "supset", "\u2283");
		add(entities, "supseteq", "\u2287");
		add(entities, "supseteqq", "\u2AC6");
		add(entities, "supsetneq", "\u228B");
		add(entities, "supsetneqq", "\u2ACC");
		add(entities, "supsim", "\u2AC8");
		add(entities, "supsub", "\u2AD4");
		add(entities, "supsup", "\u2AD6");
		add(entities, "swarhk", "\u2926");
		add(entities, "swArr", "\u21D9");
		add(entities, "swarr", "\u2199");
		add(entities, "swarrow", "\u2199");
		add(entities, "swnwar", "\u292A");
		add(entities, "szlig", "\u00DF");
		add(entities, "Tab", "\t"); // u0009
		add(entities, "target", "\u2316");
		add(entities, "Tau", "\u03A4");
		add(entities, "tau", "\u03C4");
		add(entities, "tbrk", "\u23B4");
		add(entities, "Tcaron", "\u0164");
		add(entities, "tcaron", "\u0165");
		add(entities, "Tcedil", "\u0162");
		add(entities, "tcedil", "\u0163");
		add(entities, "Tcy", "\u0422");
		add(entities, "tcy", "\u0442");
		add(entities, "tdot", "\u20DB");
		add(entities, "telrec", "\u2315");
		add(entities, "Tfr", "\uD835\uDD17");
		add(entities, "tfr", "\uD835\uDD31");
		add(entities, "there4", "\u2234");
		add(entities, "Therefore", "\u2234");
		add(entities, "therefore", "\u2234");
		add(entities, "Theta", "\u0398");
		add(entities, "theta", "\u03B8");
		add(entities, "thetasym", "\u03D1");
		add(entities, "thetav", "\u03D1");
		add(entities, "thickapprox", "\u2248");
		add(entities, "thicksim", "\u223C");
		add(entities, "ThickSpace", "\u205F\u200A");
		add(entities, "thinsp", "\u2009");
		add(entities, "ThinSpace", "\u2009");
		add(entities, "thkap", "\u2248");
		add(entities, "thksim", "\u223C");
		add(entities, "THORN", "\u00DE");
		add(entities, "THORN", "\u00DE");
		add(entities, "thorn", "\u00FE");
		add(entities, "thorn", "\u00FE");
		add(entities, "Tilde", "\u223C");
		add(entities, "tilde", "\u02DC");
		add(entities, "TildeEqual", "\u2243");
		add(entities, "TildeFullEqual", "\u2245");
		add(entities, "TildeTilde", "\u2248");
		add(entities, "times", "\u00D7");
		add(entities, "times", "\u00D7");
		add(entities, "timesb", "\u22A0");
		add(entities, "timesbar", "\u2A31");
		add(entities, "timesd", "\u2A30");
		add(entities, "tint", "\u222D");
		add(entities, "toea", "\u2928");
		add(entities, "top", "\u22A4");
		add(entities, "topbot", "\u2336");
		add(entities, "topcir", "\u2AF1");
		add(entities, "Topf", "\uD835\uDD4B");
		add(entities, "topf", "\uD835\uDD65");
		add(entities, "topfork", "\u2ADA");
		add(entities, "tosa", "\u2929");
		add(entities, "tprime", "\u2034");
		add(entities, "TRADE", "\u2122");
		add(entities, "trade", "\u2122");
		add(entities, "triangle", "\u25B5");
		add(entities, "triangledown", "\u25BF");
		add(entities, "triangleleft", "\u25C3");
		add(entities, "trianglelefteq", "\u22B4");
		add(entities, "triangleq", "\u225C");
		add(entities, "triangleright", "\u25B9");
		add(entities, "trianglerighteq", "\u22B5");
		add(entities, "tridot", "\u25EC");
		add(entities, "trie", "\u225C");
		add(entities, "triminus", "\u2A3A");
		add(entities, "TripleDot", "\u20DB");
		add(entities, "triplus", "\u2A39");
		add(entities, "trisb", "\u29CD");
		add(entities, "tritime", "\u2A3B");
		add(entities, "trpezium", "\u23E2");
		add(entities, "Tscr", "\uD835\uDCAF");
		add(entities, "tscr", "\uD835\uDCC9");
		add(entities, "TScy", "\u0426");
		add(entities, "tscy", "\u0446");
		add(entities, "TSHcy", "\u040B");
		add(entities, "tshcy", "\u045B");
		add(entities, "Tstrok", "\u0166");
		add(entities, "tstrok", "\u0167");
		add(entities, "twixt", "\u226C");
		add(entities, "twoheadleftarrow", "\u219E");
		add(entities, "twoheadrightarrow", "\u21A0");
		add(entities, "Uacute", "\u00DA");
		add(entities, "Uacute", "\u00DA");
		add(entities, "uacute", "\u00FA");
		add(entities, "uacute", "\u00FA");
		add(entities, "Uarr", "\u219F");
		add(entities, "uArr", "\u21D1");
		add(entities, "uarr", "\u2191");
		add(entities, "Uarrocir", "\u2949");
		add(entities, "Ubrcy", "\u040E");
		add(entities, "ubrcy", "\u045E");
		add(entities, "Ubreve", "\u016C");
		add(entities, "ubreve", "\u016D");
		add(entities, "Ucirc", "\u00DB");
		add(entities, "Ucirc", "\u00DB");
		add(entities, "ucirc", "\u00FB");
		add(entities, "ucirc", "\u00FB");
		add(entities, "Ucy", "\u0423");
		add(entities, "ucy", "\u0443");
		add(entities, "udarr", "\u21C5");
		add(entities, "Udblac", "\u0170");
		add(entities, "udblac", "\u0171");
		add(entities, "udhar", "\u296E");
		add(entities, "ufisht", "\u297E");
		add(entities, "Ufr", "\uD835\uDD18");
		add(entities, "ufr", "\uD835\uDD32");
		add(entities, "Ugrave", "\u00D9");
		add(entities, "Ugrave", "\u00D9");
		add(entities, "ugrave", "\u00F9");
		add(entities, "ugrave", "\u00F9");
		add(entities, "uHar", "\u2963");
		add(entities, "uharl", "\u21BF");
		add(entities, "uharr", "\u21BE");
		add(entities, "uhblk", "\u2580");
		add(entities, "ulcorn", "\u231C");
		add(entities, "ulcorner", "\u231C");
		add(entities, "ulcrop", "\u230F");
		add(entities, "ultri", "\u25F8");
		add(entities, "Umacr", "\u016A");
		add(entities, "umacr", "\u016B");
		add(entities, "uml", "\u00A8");
		add(entities, "uml", "\u00A8");
		add(entities, "UnderBar", "\u005F");
		add(entities, "UnderBrace", "\u23DF");
		add(entities, "UnderBracket", "\u23B5");
		add(entities, "UnderParenthesis", "\u23DD");
		add(entities, "Union", "\u22C3");
		add(entities, "UnionPlus", "\u228E");
		add(entities, "Uogon", "\u0172");
		add(entities, "uogon", "\u0173");
		add(entities, "Uopf", "\uD835\uDD4C");
		add(entities, "uopf", "\uD835\uDD66");
		add(entities, "UpArrow", "\u2191");
		add(entities, "Uparrow", "\u21D1");
		add(entities, "uparrow", "\u2191");
		add(entities, "UpArrowBar", "\u2912");
		add(entities, "UpArrowDownArrow", "\u21C5");
		add(entities, "UpDownArrow", "\u2195");
		add(entities, "Updownarrow", "\u21D5");
		add(entities, "updownarrow", "\u2195");
		add(entities, "UpEquilibrium", "\u296E");
		add(entities, "upharpoonleft", "\u21BF");
		add(entities, "upharpoonright", "\u21BE");
		add(entities, "uplus", "\u228E");
		add(entities, "UpperLeftArrow", "\u2196");
		add(entities, "UpperRightArrow", "\u2197");
		add(entities, "Upsi", "\u03D2");
		add(entities, "upsi", "\u03C5");
		add(entities, "upsih", "\u03D2");
		add(entities, "Upsilon", "\u03A5");
		add(entities, "upsilon", "\u03C5");
		add(entities, "UpTee", "\u22A5");
		add(entities, "UpTeeArrow", "\u21A5");
		add(entities, "upuparrows", "\u21C8");
		add(entities, "urcorn", "\u231D");
		add(entities, "urcorner", "\u231D");
		add(entities, "urcrop", "\u230E");
		add(entities, "Uring", "\u016E");
		add(entities, "uring", "\u016F");
		add(entities, "urtri", "\u25F9");
		add(entities, "Uscr", "\uD835\uDCB0");
		add(entities, "uscr", "\uD835\uDCCA");
		add(entities, "utdot", "\u22F0");
		add(entities, "Utilde", "\u0168");
		add(entities, "utilde", "\u0169");
		add(entities, "utri", "\u25B5");
		add(entities, "utrif", "\u25B4");
		add(entities, "uuarr", "\u21C8");
		add(entities, "Uuml", "\u00DC");
		add(entities, "Uuml", "\u00DC");
		add(entities, "uuml", "\u00FC");
		add(entities, "uuml", "\u00FC");
		add(entities, "uwangle", "\u29A7");
		add(entities, "vangrt", "\u299C");
		add(entities, "varepsilon", "\u03F5");
		add(entities, "varkappa", "\u03F0");
		add(entities, "varnothing", "\u2205");
		add(entities, "varphi", "\u03D5");
		add(entities, "varpi", "\u03D6");
		add(entities, "varpropto", "\u221D");
		add(entities, "vArr", "\u21D5");
		add(entities, "varr", "\u2195");
		add(entities, "varrho", "\u03F1");
		add(entities, "varsigma", "\u03C2");
		add(entities, "varsubsetneq", "\u228A\uFE00");
		add(entities, "varsubsetneqq", "\u2ACB\uFE00");
		add(entities, "varsupsetneq", "\u228B\uFE00");
		add(entities, "varsupsetneqq", "\u2ACC\uFE00");
		add(entities, "vartheta", "\u03D1");
		add(entities, "vartriangleleft", "\u22B2");
		add(entities, "vartriangleright", "\u22B3");
		add(entities, "Vbar", "\u2AEB");
		add(entities, "vBar", "\u2AE8");
		add(entities, "vBarv", "\u2AE9");
		add(entities, "Vcy", "\u0412");
		add(entities, "vcy", "\u0432");
		add(entities, "VDash", "\u22AB");
		add(entities, "Vdash", "\u22A9");
		add(entities, "vDash", "\u22A8");
		add(entities, "vdash", "\u22A2");
		add(entities, "Vdashl", "\u2AE6");
		add(entities, "Vee", "\u22C1");
		add(entities, "vee", "\u2228");
		add(entities, "veebar", "\u22BB");
		add(entities, "veeeq", "\u225A");
		add(entities, "vellip", "\u22EE");
		add(entities, "Verbar", "\u2016");
		add(entities, "verbar", "\u007C");
		add(entities, "Vert", "\u2016");
		add(entities, "vert", "\u007C");
		add(entities, "VerticalBar", "\u2223");
		add(entities, "VerticalLine", "\u007C");
		add(entities, "VerticalSeparator", "\u2758");
		add(entities, "VerticalTilde", "\u2240");
		add(entities, "VeryThinSpace", "\u200A");
		add(entities, "Vfr", "\uD835\uDD19");
		add(entities, "vfr", "\uD835\uDD33");
		add(entities, "vltri", "\u22B2");
		add(entities, "vnsub", "\u2282\u20D2");
		add(entities, "vnsup", "\u2283\u20D2");
		add(entities, "Vopf", "\uD835\uDD4D");
		add(entities, "vopf", "\uD835\uDD67");
		add(entities, "vprop", "\u221D");
		add(entities, "vrtri", "\u22B3");
		add(entities, "Vscr", "\uD835\uDCB1");
		add(entities, "vscr", "\uD835\uDCCB");
		add(entities, "vsubnE", "\u2ACB\uFE00");
		add(entities, "vsubne", "\u228A\uFE00");
		add(entities, "vsupnE", "\u2ACC\uFE00");
		add(entities, "vsupne", "\u228B\uFE00");
		add(entities, "Vvdash", "\u22AA");
		add(entities, "vzigzag", "\u299A");
		add(entities, "Wcirc", "\u0174");
		add(entities, "wcirc", "\u0175");
		add(entities, "wedbar", "\u2A5F");
		add(entities, "Wedge", "\u22C0");
		add(entities, "wedge", "\u2227");
		add(entities, "wedgeq", "\u2259");
		add(entities, "weierp", "\u2118");
		add(entities, "Wfr", "\uD835\uDD1A");
		add(entities, "wfr", "\uD835\uDD34");
		add(entities, "Wopf", "\uD835\uDD4E");
		add(entities, "wopf", "\uD835\uDD68");
		add(entities, "wp", "\u2118");
		add(entities, "wr", "\u2240");
		add(entities, "wreath", "\u2240");
		add(entities, "Wscr", "\uD835\uDCB2");
		add(entities, "wscr", "\uD835\uDCCC");
		add(entities, "xcap", "\u22C2");
		add(entities, "xcirc", "\u25EF");
		add(entities, "xcup", "\u22C3");
		add(entities, "xdtri", "\u25BD");
		add(entities, "Xfr", "\uD835\uDD1B");
		add(entities, "xfr", "\uD835\uDD35");
		add(entities, "xhArr", "\u27FA");
		add(entities, "xharr", "\u27F7");
		add(entities, "Xi", "\u039E");
		add(entities, "xi", "\u03BE");
		add(entities, "xlArr", "\u27F8");
		add(entities, "xlarr", "\u27F5");
		add(entities, "xmap", "\u27FC");
		add(entities, "xnis", "\u22FB");
		add(entities, "xodot", "\u2A00");
		add(entities, "Xopf", "\uD835\uDD4F");
		add(entities, "xopf", "\uD835\uDD69");
		add(entities, "xoplus", "\u2A01");
		add(entities, "xotime", "\u2A02");
		add(entities, "xrArr", "\u27F9");
		add(entities, "xrarr", "\u27F6");
		add(entities, "Xscr", "\uD835\uDCB3");
		add(entities, "xscr", "\uD835\uDCCD");
		add(entities, "xsqcup", "\u2A06");
		add(entities, "xuplus", "\u2A04");
		add(entities, "xutri", "\u25B3");
		add(entities, "xvee", "\u22C1");
		add(entities, "xwedge", "\u22C0");
		add(entities, "Yacute", "\u00DD");
		add(entities, "Yacute", "\u00DD");
		add(entities, "yacute", "\u00FD");
		add(entities, "yacute", "\u00FD");
		add(entities, "YAcy", "\u042F");
		add(entities, "yacy", "\u044F");
		add(entities, "Ycirc", "\u0176");
		add(entities, "ycirc", "\u0177");
		add(entities, "Ycy", "\u042B");
		add(entities, "ycy", "\u044B");
		add(entities, "yen", "\u00A5");
		add(entities, "yen", "\u00A5");
		add(entities, "Yfr", "\uD835\uDD1C");
		add(entities, "yfr", "\uD835\uDD36");
		add(entities, "YIcy", "\u0407");
		add(entities, "yicy", "\u0457");
		add(entities, "Yopf", "\uD835\uDD50");
		add(entities, "yopf", "\uD835\uDD6A");
		add(entities, "Yscr", "\uD835\uDCB4");
		add(entities, "yscr", "\uD835\uDCCE");
		add(entities, "YUcy", "\u042E");
		add(entities, "yucy", "\u044E");
		add(entities, "Yuml", "\u0178");
		add(entities, "yuml", "\u00FF");
		add(entities, "yuml", "\u00FF");
		add(entities, "Zacute", "\u0179");
		add(entities, "zacute", "\u017A");
		add(entities, "Zcaron", "\u017D");
		add(entities, "zcaron", "\u017E");
		add(entities, "Zcy", "\u0417");
		add(entities, "zcy", "\u0437");
		add(entities, "Zdot", "\u017B");
		add(entities, "zdot", "\u017C");
		add(entities, "zeetrf", "\u2128");
		add(entities, "ZeroWidthSpace", "\u200B");
		add(entities, "Zeta", "\u0396");
		add(entities, "zeta", "\u03B6");
		add(entities, "Zfr", "\u2128");
		add(entities, "zfr", "\uD835\uDD37");
		add(entities, "ZHcy", "\u0416");
		add(entities, "zhcy", "\u0436");
		add(entities, "zigrarr", "\u21DD");
		add(entities, "Zopf", "\u2124");
		add(entities, "zopf", "\uD835\uDD6B");
		add(entities, "Zscr", "\uD835\uDCB5");
		add(entities, "zscr", "\uD835\uDCCF");
		add(entities, "zwj", "\u200D");
		add(entities, "zwnj", "\u200C");
		
		NAMED_ENTITIES= entities;
	}
	
	private static void add(final Map<String, Entity> entities, final String name, final String string) {
		entities.put(name, new Entity(name, string));
	}
	
	
	public static String getEntityReference(final String entity) {
		final int end= entity.length();
		if (end >= 2 && entity.charAt(0) == '&') {
			return entity.substring(1, (entity.charAt(end - 1) == ';') ? (end - 1) : end);
		}
		return entity;
	}
	
	public static @Nullable Entity getNamedEntity(final String name) {
		return NAMED_ENTITIES.get(name);
	}
	
	public static @Nullable String resolveEntity(final String reference) throws IllegalArgumentException {
		final int length= reference.length();
		if (length > 0) {
			if (reference.charAt(0) == '#') {
				final int codePoint;
				if (length > 1
						&& (reference.charAt(1) == 'x' || reference.charAt(1) == 'X') ) {
					codePoint= Integer.parseInt(reference.substring(2), 16);
				}
				else {
					codePoint= Integer.parseInt(reference.substring(1), 10);
				}
				return new String(Character.toChars(codePoint));
			}
			{	final Entity entity= NAMED_ENTITIES.get(reference);
				if (entity != null) {
					return entity.getString();
				}
			}
		}
		return null;
	}
	
}
